__all__ = [
    "DataConnection",
    "S3Connection",
    "ConnectionAsset",
    "S3Location",
    "FSLocation",
    "AssetLocation",
    "CP4DAssetLocation",
    "WMLSAssetLocation",
    "WSDAssetLocation",
    "CloudAssetLocation",
    "DeploymentOutputAssetLocation",
    "NFSConnection",
    "NFSLocation",
    'ConnectionAssetLocation',
    "DatabaseLocation",
    "ContainerLocation"
]

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import io
import os
import uuid
import copy
import sys
from copy import deepcopy
from typing import Union, Tuple, List, TYPE_CHECKING, Optional
from warnings import warn

from ibm_boto3 import resource
from ibm_botocore.client import ClientError
from pandas import DataFrame
import pandas as pd
import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, DataConnectionTypes
from ibm_watson_machine_learning.utils.autoai.errors import (
    MissingAutoPipelinesParameters, UseWMLClient, MissingCOSStudioConnection, MissingProjectLib,
    HoldoutSplitNotSupported, InvalidCOSCredentials, MissingLocalAsset, InvalidIdType, NotWSDEnvironment,
    NotExistingCOSResource, InvalidDataAsset, CannotReadSavedRemoteDataBeforeFit, NoAutomatedHoldoutSplit
)

import numpy as np
from ibm_watson_machine_learning.utils.autoai.utils import all_logging_disabled, try_import_autoai_libs, \
    try_import_autoai_ts_libs
from ibm_watson_machine_learning.utils.autoai.watson_studio import get_project
from ibm_watson_machine_learning.data_loaders.datasets.experiment import DEFAULT_SAMPLING_TYPE, DEFAULT_SAMPLE_SIZE_LIMIT
from ibm_watson_machine_learning.wml_client_error import MissingValue, ApiRequestFailure, WMLClientError
from ibm_watson_machine_learning.utils.autoai.errors import ContainerTypeNotSupported
from .base_connection import BaseConnection
from .base_data_connection import BaseDataConnection
from .base_location import BaseLocation

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace


class DataConnection(BaseDataConnection):
    """Data Storage Connection class needed for WML training metadata (input data).

    :param connection: connection parameters of specific type
    :type connection: NFSConnection or ConnectionAsset, optional

    :param location: required location parameters of specific type
    :type location: Union[S3Location, FSLocation, AssetLocation]

    :param data_join_node_name: name(s) for node(s):

        - `None` - data file name will be used as node name
        - str - it will became node name
        - list[str] - multiple names passed, several nodes will have the same data connection
          (used for excel files with multiple sheets)
    :type data_join_node_name:  None or str or list[str], optional

    :param data_asset_id: data asset ID if DataConnection should be pointing out to data asset
    :type data_asset_id: str, optional
    """

    def __init__(self,
                 location: Union['S3Location',
                                 'FSLocation',
                                 'AssetLocation',
                                 'CP4DAssetLocation',
                                 'WMLSAssetLocation',
                                 'WSDAssetLocation',
                                 'CloudAssetLocation',
                                 'NFSLocation',
                                 'DeploymentOutputAssetLocation',
                                 'ConnectionAssetLocation',
                                 'DatabaseLocation',
                                 'ContainerLocation'] = None,
                 connection: Optional[Union['S3Connection', 'NFSConnection', 'ConnectionAsset']] = None,
                 data_join_node_name: Union[str, List[str]] = None,
                 data_asset_id: str = None,
                 connection_asset_id: str = None,
                 **kwargs):
        if data_asset_id is None and location is None:
            raise MissingValue('location or data_asset_id', reason="Provide 'location' or 'data_asset_id'.")

        elif data_asset_id is not None and location is not None:
            raise ValueError("'data_asset_id' and 'location' cannot be specified together.")

        elif data_asset_id is not None:
            location = AssetLocation(asset_id=data_asset_id)

            if kwargs.get('model_location') is not None:
                location._model_location = kwargs['model_location']

            if kwargs.get('training_status') is not None:
                location._training_status = kwargs['training_status']

        elif connection_asset_id is not None and isinstance(location, (S3Location, DatabaseLocation, NFSLocation)):
            connection = ConnectionAsset(connection_id=connection_asset_id)

        super().__init__()

        self.connection = connection
        self.location = location

        # TODO: remove S3 implementation
        if isinstance(connection, S3Connection):
            self.type = DataConnectionTypes.S3

        elif isinstance(connection, ConnectionAsset):
            self.type = DataConnectionTypes.CA
            # note: We expect a `file_name` keyword for CA pointing to COS or NFS.
            if isinstance(self.location, (S3Location, NFSLocation)):
                self.location.file_name = self.location.path
                del self.location.path
                if isinstance(self.location, NFSLocation):
                    del self.location.id
            # --- end note

        elif isinstance(location, FSLocation):
            self.type = DataConnectionTypes.FS

        elif isinstance(location, ContainerLocation):
            self.type = DataConnectionTypes.CN

        elif isinstance(location, (AssetLocation, CP4DAssetLocation, WMLSAssetLocation, CloudAssetLocation,
                                   WSDAssetLocation, DeploymentOutputAssetLocation)):
            self.type = DataConnectionTypes.DS

        self.auto_pipeline_params = {}  # note: needed parameters for recreation of autoai holdout split
        self._wml_client = None
        self.__wml_client = None  # only for getter/setter for AssetLocation href
        self._run_id = None
        self._obm = False
        self._obm_cos_path = None
        self._test_data = False
        self._user_holdout_exists = False

        # note: make data connection id as a location path for OBM + KB
        if data_join_node_name is None:
            # TODO: remove S3 implementation
            if self.type == DataConnectionTypes.S3 or (
                    self.type == DataConnectionTypes.CA and hasattr(location, 'file_name')):
                self.id = location.get_location()

            else:
                self.id = None

        else:
            self.id = data_join_node_name
        # --- end note

    # note: client as property and setter for dynamic href creation for AssetLocation
    @property
    def _wml_client(self):
        return self.__wml_client

    @_wml_client.setter
    def _wml_client(self, var):
        self.__wml_client = var
        if isinstance(self.location, (AssetLocation, WSDAssetLocation)):
            self.location.wml_client = self.__wml_client

        if getattr(var, 'project_type', None) == 'local_git_storage':
            self.location.userfs = True

    def set_client(self, wml_client):
        """Set initialized wml client in connection to enable write/read operations with connection to service.

        :param wml_client: WML client to connect to service
        :type wml_client: APIClient

        **Example**

        .. code-block:: python

            DataConnection.set_client(wml_client)
        """
        self._wml_client = wml_client

    # --- end note

    @classmethod
    def from_studio(cls, path: str) -> List['DataConnection']:
        """Create DataConnections from the credentials stored (connected) in Watson Studio. Only for COS.

        :param path: path in COS bucket to the training dataset
        :type path: str

        :return: list with DataConnection objects
        :rtype: list[DataConnection]

        **Example**

        .. code-block:: python

            data_connections = DataConnection.from_studio(path='iris_dataset.csv')
        """
        try:
            from project_lib import Project

        except ModuleNotFoundError:
            raise MissingProjectLib("Missing project_lib package.")

        else:
            data_connections = []
            for name, value in globals().items():
                if isinstance(value, Project):
                    connections = value.get_connections()

                    if connections:
                        for connection in connections:
                            asset_id = connection['asset_id']
                            connection_details = value.get_connection(asset_id)

                            if ('url' in connection_details and 'access_key' in connection_details and
                                    'secret_key' in connection_details and 'bucket' in connection_details):
                                data_connections.append(
                                    cls(connection=ConnectionAsset(id=connection_details['id']),
                                        location=ConnectionAssetLocation(bucket=connection_details['bucket'],
                                                                         file_name=path))
                                )

            if data_connections:
                return data_connections

            else:
                raise MissingCOSStudioConnection(
                    "There is no any COS Studio connection. "
                    "Please create a COS connection from the UI and insert "
                    "the cell with project API connection (Insert project token)")

    def _subdivide_connection(self):
        if type(self.id) is str or not self.id:
            return [self]
        else:
            def cpy(new_id):
                child = copy.copy(self)
                child.id = new_id
                return child

            return [cpy(id) for id in self.id]

    def _to_dict(self) -> dict:
        """Convert DataConnection object to dictionary representation.

        :return: DataConnection dictionary representation
        :rtype: dict
        """

        if self.id and type(self.id) is list:
            raise InvalidIdType(list)

        _dict = {"type": self.type}

        # note: for OBM (id of DataConnection if an OBM node name)
        if self.id is not None:
            _dict['id'] = self.id
        # --- end note

        if self.connection is not None:
            _dict['connection'] = deepcopy(self.connection.to_dict())
        else:
            _dict['connection'] = {}

        try:
            _dict['location'] = deepcopy(self.location.to_dict())

        except AttributeError:
            _dict['location'] = {}

        # note: convert userfs to string - training service requires it as string
        if getattr(self.location, 'userfs', False):
            _dict['location']['userfs'] = str(getattr(self.location, 'userfs', False)).lower()
        # end note

        return _dict

    def __repr__(self):
        return str(self._to_dict())

    def __str__(self):
        return str(self._to_dict())

    @classmethod
    def _from_dict(cls, _dict: dict) -> 'DataConnection':
        """Create a DataConnection object from dictionary.

        :param _dict: a dictionary data structure with information about data connection reference
        :type _dict: dict

        :return: DataConnection object
        :rtype: DataConnection
        """
        # TODO: remove S3 implementation
        if _dict['type'] == DataConnectionTypes.S3:
            warn(message="S3 DataConnection is deprecated! Please use data_asset_id instead.")

            data_connection: 'DataConnection' = cls(
                connection=S3Connection(
                    access_key_id=_dict['connection']['access_key_id'],
                    secret_access_key=_dict['connection']['secret_access_key'],
                    endpoint_url=_dict['connection']['endpoint_url']
                ),
                location=S3Location(
                    bucket=_dict['location']['bucket'],
                    path=_dict['location']['path']
                )
            )
        elif _dict['type'] == DataConnectionTypes.FS:
            data_connection: 'DataConnection' = cls(
                location=FSLocation._set_path(path=_dict['location']['path'])
            )
        elif _dict['type'] == DataConnectionTypes.CA:
            if _dict['location'].get('file_name') is not None and _dict['location'].get('bucket'):
                data_connection: 'DataConnection' = cls(
                    connection_asset_id=_dict['connection']['id'],
                    location=S3Location(
                        bucket=_dict['location']['bucket'],
                        path=_dict['location']['file_name']
                    )
                )

            elif _dict['location'].get('path') is not None and _dict['location'].get('bucket'):
                data_connection: 'DataConnection' = cls(
                    connection_asset_id=_dict['connection']['id'],
                    location=S3Location(
                        bucket=_dict['location']['bucket'],
                        path=_dict['location']['path']
                    )
                )

            elif _dict['location'].get('schema_name') and _dict['location'].get('table_name'):
                data_connection: 'DataConnection' = cls(
                    connection_asset_id=_dict['connection']['id'],
                    location=DatabaseLocation(schema_name=_dict['location']['schema_name'],
                                              table_name=_dict['location']['table_name'])
                )

            else:
                data_connection: 'DataConnection' = cls(
                    connection=NFSConnection(asset_id=_dict['connection']['asset_id']),
                    location=NFSLocation(path=_dict['location']['path'])
                )
        elif _dict['type'] == DataConnectionTypes.CN:
            data_connection: 'DataConnection' = cls(
                location=ContainerLocation(path=_dict['location']['path'])
            )

        else:
            data_connection: 'DataConnection' = cls(
                location=AssetLocation._set_path(href=_dict['location']['href'])
            )

        if _dict.get('id'):
            data_connection.id = _dict['id']

        if _dict['location'].get('userfs'):
            if str(_dict['location'].get('userfs', 'false')).lower() in ['true', '1']:
                data_connection.location.userfs = True
            else:
                data_connection.location.userfs = False

        return data_connection

    def _recreate_holdout(
            self,
            data: 'DataFrame',
            with_holdout_split: bool = True
    ) -> Union[Tuple['DataFrame', 'DataFrame'], Tuple['DataFrame', 'DataFrame', 'DataFrame', 'DataFrame']]:
        """This method tries to recreate holdout data."""

        if self.auto_pipeline_params.get('prediction_columns') is not None:
            # timeseries
            try_import_autoai_ts_libs()
            from autoai_ts_libs.utils.holdout_utils import make_holdout_split

            # Note: When lookback window is auto detected there is need to get the detected value from training details
            if self.auto_pipeline_params.get('lookback_window') == -1 or self.auto_pipeline_params.get('lookback_window') is None:
                ts_metrics = self._wml_client.training.get_details(self.auto_pipeline_params.get('run_id'), _internal=True)['entity']['status']['metrics']
                final_ts_state_name = "after_final_pipelines_generation"


                for metric in ts_metrics:
                    if metric['context']['intermediate_model']['process'] == final_ts_state_name:
                        self.auto_pipeline_params['lookback_window'] = metric['context']['timeseries']['lookback_window']
                        break

            # Note: imputation is not supported
            X_train, X_holdout, y_train, y_holdout, _, _, _, _ = make_holdout_split(
                dataset=data,
                target_columns=self.auto_pipeline_params.get('prediction_columns'),
                learning_type="forecasting",
                test_size=self.auto_pipeline_params.get('holdout_size'),
                lookback_window=self.auto_pipeline_params.get('lookback_window'),
                feature_columns=self.auto_pipeline_params.get('feature_columns'),
                timestamp_column=self.auto_pipeline_params.get('timestamp_column_name'),
                # n_jobs=None,
                # tshirt_size=None,
                return_only_holdout=False
            )

            X_columns = self.auto_pipeline_params.get('feature_columns') if self.auto_pipeline_params.get('feature_columns') else self.auto_pipeline_params['prediction_columns']

            X_train = DataFrame(X_train, columns=X_columns)
            X_holdout = DataFrame(X_holdout, columns=X_columns)
            y_train = DataFrame(y_train, columns=self.auto_pipeline_params['prediction_columns'])
            y_holdout = DataFrame(y_holdout, columns=self.auto_pipeline_params['prediction_columns'])

            return X_train, X_holdout, y_train, y_holdout
        elif self.auto_pipeline_params.get('feature_columns') is not None:
            # timeseries anomaly detection
            try_import_autoai_ts_libs()
            from autoai_ts_libs.utils.holdout_utils import make_holdout_split
            from autoai_ts_libs.utils.constants import LEARNING_TYPE_TIMESERIES_ANOMALY_PREDICTION

            # Note: imputation is not supported
            X_train, X_holdout, y_train, y_holdout, _, _, _, _ = make_holdout_split(
                dataset=data,
                learning_type=LEARNING_TYPE_TIMESERIES_ANOMALY_PREDICTION,
                test_size=self.auto_pipeline_params.get('holdout_size'),
                # lookback_window=self.auto_pipeline_params.get('lookback_window'),
                feature_columns=self.auto_pipeline_params.get('feature_columns'),
                timestamp_column=self.auto_pipeline_params.get('timestamp_column_name'),
                # n_jobs=None,
                # tshirt_size=None,
                return_only_holdout=False
            )

            X_columns = self.auto_pipeline_params['feature_columns']
            y_column = ['anomaly_label']

            X_train = DataFrame(X_train, columns=X_columns)
            X_holdout = DataFrame(X_holdout, columns=X_columns)
            y_train = DataFrame(y_train, columns=y_column)
            y_holdout = DataFrame(y_holdout, columns=y_column)

            return X_train, X_holdout, y_train, y_holdout

        else:
            if sys.version_info >= (3, 10):
                try_import_autoai_libs(minimum_version='1.14.0')
            else:
                try_import_autoai_libs(minimum_version='1.12.14')

            from autoai_libs.utils.holdout_utils import make_holdout_split, numpy_split_on_target_values
            from autoai_libs.utils.sampling_utils import numpy_sample_rows

            data.replace([np.inf, -np.inf], np.nan, inplace=True)
            data.drop_duplicates(inplace=True)
            data.dropna(subset=[self.auto_pipeline_params['prediction_column']], inplace=True)
            dfy = data[self.auto_pipeline_params['prediction_column']]
            data.drop(columns=[self.auto_pipeline_params['prediction_column']], inplace=True)

            y_column = [self.auto_pipeline_params['prediction_column']]
            X_columns = data.columns

            if self._test_data or not with_holdout_split:
                return data, dfy

            else:
                ############################
                #   REMOVE MISSING ROWS    #
                from autoai_libs.utils.holdout_utils import numpy_remove_missing_target_rows
                # Remove (and save) the rows of X and y for which the target variable has missing values
                data, dfy, _, _, _, _ = numpy_remove_missing_target_rows(
                    y=dfy, X=data
                )
                #   End of REMOVE MISSING ROWS    #
                ###################################

                #################
                #   SAMPLING    #
                # Get a sample of the rows if requested and applicable
                # (check for sampling is performed inside this function)
                try:
                    data, dfy, _ = numpy_sample_rows(
                        X=data,
                        y=dfy,
                        train_sample_rows_test_size=self.auto_pipeline_params['train_sample_rows_test_size'],
                        learning_type=self.auto_pipeline_params['prediction_type'],
                        return_sampled_indices=True
                    )

                # Note: we have a silent error here (the old core behaviour)
                # sampling is not performed as 'train_sample_rows_test_size' is bigger than data rows count
                # TODO: can we throw an error instead?
                except ValueError as e:
                    if 'between' in str(e):
                        pass

                    else:
                        raise e
                #   End of SAMPLING    #
                ########################

                # Perform holdout split
                X_train, X_holdout, y_train, y_holdout, _, _ = make_holdout_split(
                    x=data,
                    y=dfy,
                    learning_type=self.auto_pipeline_params['prediction_type'],
                    fairness_info=self.auto_pipeline_params.get('fairness_info', None),
                    test_size=self.auto_pipeline_params['holdout_size'],
                    return_only_holdout=False
                )

                X_train = DataFrame(X_train, columns=X_columns)
                X_holdout = DataFrame(X_holdout, columns=X_columns)
                y_train = DataFrame(y_train, columns=y_column)
                y_holdout = DataFrame(y_holdout, columns=y_column)

                return X_train, X_holdout, y_train, y_holdout

    def read(self,
             with_holdout_split: bool = False,
             csv_separator: str = ',',
             excel_sheet: Union[str, int] = None,
             encoding: Optional[str] = 'utf-8',
             raw: Optional[bool] = False,
             binary: Optional[bool] = False,
             read_to_file: Optional[str] = None,
             number_of_batch_rows: Optional[int] = None,
             sampling_type: Optional[str] = None,
             sample_size_limit: Optional[int] = None,
             sample_rows_limit: Optional[int] = None,
             sample_percentage_limit: Optional[float] = None,
             **kwargs) -> Union['DataFrame', Tuple['DataFrame', 'DataFrame'], bytes]:
        """Download dataset stored in remote data storage. Returns batch up to 1GB.

        :param with_holdout_split: if `True`, data will be split to train and holdout dataset as it was by AutoAI
        :type with_holdout_split: bool, optional

        :param csv_separator: separator / delimiter for CSV file
        :type csv_separator: str, optional

        :param excel_sheet: excel file sheet name to use, only use when xlsx file is an input,
            support for number of the sheet is deprecated
        :type excel_sheet: str, optional

        :param encoding: encoding type of the CSV
        :type encoding: str, optional

        :param raw: if `False` there wil be applied simple data preprocessing (the same as in the backend),
            if `True`, data will be not preprocessed
        :type raw: bool, optional

        :param binary: indicates to retrieve data in binary mode, the result will be a python binary type variable
        :type binary: bool, optional

        :param read_to_file: stream read data to file under path specified as value of this parameter,
            use this parameter to prevent keeping data in-memory
        :type read_to_file: str, optional

        :param number_of_batch_rows: number of rows to read in each batch when reading from flight connection
        :type number_of_batch_rows: int, optional

        :param sampling_type: a sampling strategy how to read the data
        :type sampling_type: str, optional

        :param sample_size_limit: upper limit for overall data that should be downloaded in bytes, default: 1 GB
        :type sample_size_limit: int, optional

        :param sample_rows_limit: upper limit for overall data that should be downloaded in number of rows
        :type sample_rows_limit: int, optional

        :param sample_percentage_limit: upper limit for overall data that should be downloaded
            in percent of all dataset, this parameter is ignored, when `sampling_type` parameter is set
            to `first_n_records`, must be a float number between 0 and 1
        :type sample_percentage_limit: float, optional

        .. note::

            If more than one of: `sample_size_limit`, `sample_rows_limit`, `sample_percentage_limit` are set,
            then downloaded data is limited to the lowest threshold.

        :return: one of:

            - pandas.DataFrame contains dataset from remote data storage : Xy_train
            - Tuple[pandas.DataFrame, pandas.DataFrame, pandas.DataFrame, pandas.DataFrame] : X_train, X_holdout, y_train, y_holdout
            - Tuple[pandas.DataFrame, pandas.DataFrame] : X_test, y_test containing training data and holdout data from
              remote storage
            - bytes object, auto holdout split from backend (only train data provided)

        **Examples**

        .. code-block:: python

            train_data_connections = optimizer.get_data_connections()

            data = train_data_connections[0].read() # all train data

            # or

            X_train, X_holdout, y_train, y_holdout = train_data_connections[0].read(with_holdout_split=True) # train and holdout data

        User provided train and test data:

        .. code-block:: python

            optimizer.fit(training_data_reference=[DataConnection],
                          training_results_reference=DataConnection,
                          test_data_reference=DataConnection)

            test_data_connection = optimizer.get_test_data_connections()
            X_test, y_test = test_data_connection.read() # only holdout data

            # and

            train_data_connections = optimizer.get_data_connections()
            data = train_connections[0].read() # only train data
        """

        # enables flight automatically for CP4D 4.0.x, 4.5.x, for 3.0 and 3.5 we do not have a flight service there
        try:
            use_flight = kwargs.get(
                'use_flight',
                bool((self._wml_client is not None or 'USER_ACCESS_TOKEN' in os.environ or 'RUNTIME_ENV_ACCESS_TOKEN_FILE' in os.environ) and (
                        self._wml_client.ICP_40 or self._wml_client.ICP_45 or self._wml_client.ICP_46)))
        except:
            use_flight = False

        return_data_as_iterator = kwargs.get('return_data_as_iterator', False)
        sampling_type = sampling_type if sampling_type is not None else DEFAULT_SAMPLING_TYPE
        enable_sampling = kwargs.get('enable_sampling', True)
        read_normal_data_asset_without_flight = kwargs.get('read_normal_data_asset_without_flight', False)
        total_size_limit = sample_size_limit if sample_size_limit is not None else kwargs.get('total_size_limit', DEFAULT_SAMPLE_SIZE_LIMIT)
        total_nrows_limit = sample_rows_limit
        total_percentage_limit = sample_percentage_limit if sample_percentage_limit is not None else 1.0

        # Deprecation of excel_sheet as number:
        if isinstance(excel_sheet, int):
            warn(
                message="Support for excel sheet as number of the sheet (int) is deprecated! Please set excel sheet with name of the sheet.")

        flight_parameters = kwargs.get('flight_parameters', {})
        impersonate_header = kwargs.get('impersonate_header', None)

        if with_holdout_split and self._user_holdout_exists:  # when this connection is training one
            raise NoAutomatedHoldoutSplit(reason="Experiment was run based on user defined holdout dataset.")

        # note: experiment metadata is used only in autogen notebooks
        experiment_metadata = kwargs.get('experiment_metadata')
        # note: process subsampling stats flag
        _return_subsampling_stats = kwargs.get("_return_subsampling_stats", False)

        if experiment_metadata is not None:
            self.auto_pipeline_params['train_sample_rows_test_size'] = experiment_metadata.get(
                'train_sample_rows_test_size')
            self.auto_pipeline_params['prediction_column'] = experiment_metadata.get('prediction_column')
            self.auto_pipeline_params['prediction_columns'] = experiment_metadata.get('prediction_columns')
            self.auto_pipeline_params['holdout_size'] = experiment_metadata.get('holdout_size')
            self.auto_pipeline_params['prediction_type'] = experiment_metadata['prediction_type']
            self.auto_pipeline_params['fairness_info'] = experiment_metadata.get('fairness_info')
            self.auto_pipeline_params['lookback_window'] = experiment_metadata.get('lookback_window')
            self.auto_pipeline_params['timestamp_column_name'] = experiment_metadata.get('timestamp_column_name')
            self.auto_pipeline_params['feature_columns'] = experiment_metadata.get('feature_columns')

            # note: check for cloud
            if 'training_result_reference' in experiment_metadata:
                if isinstance(experiment_metadata['training_result_reference'].location, (S3Location, AssetLocation)):
                    run_id = experiment_metadata['training_result_reference'].location._training_status.split('/')[-2]
                # WMLS
                else:
                    run_id = experiment_metadata['training_result_reference'].location.path.split('/')[-3]
                self.auto_pipeline_params['run_id'] = run_id

            if self._test_data:
                csv_separator = experiment_metadata.get('test_data_csv_separator', csv_separator)
                excel_sheet = experiment_metadata.get('test_data_excel_sheet', excel_sheet)
                encoding = experiment_metadata.get('test_data_encoding', encoding)

            else:
                csv_separator = experiment_metadata.get('csv_separator', csv_separator)
                excel_sheet = experiment_metadata.get('excel_sheet', excel_sheet)
                encoding = experiment_metadata.get('encoding', encoding)

        if self.type == DataConnectionTypes.DS or self.type == DataConnectionTypes.CA:
            if self._wml_client is None:
                try:
                    from project_lib import Project

                except ModuleNotFoundError:
                    raise ConnectionError(
                        "This functionality can be run only on Watson Studio or with wml_client passed to connection. "
                        "Please initialize WML client using `DataConnection.set_client(wml_client)` function "
                        "to be able to use this functionality.")

        if (with_holdout_split or self._test_data) and not self.auto_pipeline_params.get('prediction_type', False):
            raise MissingAutoPipelinesParameters(
                self.auto_pipeline_params,
                reason=f"To be able to recreate an original holdout split, you need to schedule a training job or "
                       f"if you are using historical runs, just call historical_optimizer.get_data_connections()")

        # note: allow to read data at any time
        elif (('csv_separator' not in self.auto_pipeline_params and 'encoding' not in self.auto_pipeline_params)
              or csv_separator != ',' or encoding != 'utf-8'):
            self.auto_pipeline_params['csv_separator'] = csv_separator
            self.auto_pipeline_params['encoding'] = encoding
        # --- end note
        # note: excel_sheet in params only if it is not None (not specified):
        if excel_sheet:
            self.auto_pipeline_params['excel_sheet'] = excel_sheet
        # --- end note

        # note: set default quote character for flight (later applicable only for csv files stored in S3)
        self.auto_pipeline_params['quote_character'] = 'double_quote'
        # --- end note

        data = DataFrame()

        headers = None
        if self._wml_client is None:
            token = self._get_token_from_environment()
            if token is not None:
                headers = {'Authorization': f'Bearer {token}'}
        elif impersonate_header is not None:
            headers = self._wml_client._get_headers()

            headers['impersonate'] = impersonate_header
        if self.type == DataConnectionTypes.S3:
            raise ConnectionError(
                f"S3 DataConnection is deprecated! Please use data_asset_id instead.")

        elif self.type == DataConnectionTypes.DS:
            # workaround for reading data assets on CPD 4.5: issue #28332
            data = None
            if read_normal_data_asset_without_flight and self._wml_client and self._is_data_asset_normal():
                data = self._download_training_data_from_data_asset_storage()

            if data is None:  # read data in normal scenario with flight
                if use_flight and not self._obm:
                    from ibm_watson_machine_learning.utils.utils import is_lib_installed

                    is_lib_installed(lib_name='pyarrow', minimum_version='3.0.0', install=True)

                    from pyarrow.flight import FlightError

                    _iam_id = None
                    if headers and headers.get('impersonate'):
                        _iam_id = headers.get('impersonate', {}).get('iam_id')

                    self._wml_client._iam_id = _iam_id

                    try:
                        if self._check_if_connection_asset_is_s3():
                            # note: update flight parameters only if `connection_properties` was not set earlier
                            #       (e.x. by wml/autoi)
                            if not flight_parameters.get('connection_properties'):
                                flight_parameters = self._update_flight_parameters_with_connection_details(flight_parameters)

                        data = self._download_data_from_flight_service(data_location=self,
                                                                       binary=binary,
                                                                       read_to_file=read_to_file,
                                                                       flight_parameters=flight_parameters,
                                                                       headers=headers,
                                                                       enable_sampling=enable_sampling,
                                                                       sampling_type=sampling_type,
                                                                       number_of_batch_rows=number_of_batch_rows,
                                                                       return_data_as_iterator=return_data_as_iterator,
                                                                       _return_subsampling_stats=_return_subsampling_stats,
                                                                       total_size_limit=total_size_limit,
                                                                       total_nrows_limit=total_nrows_limit,
                                                                       total_percentage_limit=total_percentage_limit
                                                                       )
                    except (ConnectionError, FlightError, ApiRequestFailure) as download_data_error:
                        # note: try to download normal data asset either directly from cams or from mounted NFS
                        #       to keep backward compatibility
                        if self._wml_client and (self._is_data_asset_normal() or self._is_data_asset_nfs()):
                            import warnings
                            warnings.warn(str(download_data_error), Warning)
                            data = self._download_training_data_from_data_asset_storage()
                        else:
                            raise download_data_error

                # backward compatibility
                else:
                    try:
                        with all_logging_disabled():
                            if self._check_if_connection_asset_is_s3():
                                cos_client = self._init_cos_client()

                                if self._obm:
                                    data = self._download_obm_data_from_cos(cos_client=cos_client)

                                else:
                                    data = self._download_data_from_cos(cos_client=cos_client)
                            else:
                                data = self._download_training_data_from_data_asset_storage()

                    except NotImplementedError as e:
                        raise e

                    except FileNotFoundError as e:
                        raise e

                    except Exception as e:
                        # do not try Flight if we are on the cloud
                        if self._wml_client is not None:
                            if not self._wml_client.ICP:
                                raise e

                        elif os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                            raise CannotReadSavedRemoteDataBeforeFit()

                        data = self._download_data_from_flight_service(data_location=self,
                                                                       binary=binary,
                                                                       read_to_file=read_to_file,
                                                                       flight_parameters=flight_parameters,
                                                                       headers=headers,
                                                                       enable_sampling=enable_sampling,
                                                                       sampling_type=sampling_type,
                                                                       number_of_batch_rows=number_of_batch_rows,
                                                                       return_data_as_iterator=return_data_as_iterator,
                                                                       _return_subsampling_stats=_return_subsampling_stats,
                                                                       total_size_limit=total_size_limit,
                                                                       total_nrows_limit=total_nrows_limit,
                                                                       total_percentage_limit=total_percentage_limit)

        elif self.type == DataConnectionTypes.FS:

            if self._obm:
                data = self._download_obm_data_from_file_system()
            else:
                data = self._download_training_data_from_file_system()

        elif self.type == DataConnectionTypes.CA or self.type == DataConnectionTypes.CN:
            if getattr(self._wml_client, 'ICP', False) and self.type == DataConnectionTypes.CN:
                raise ContainerTypeNotSupported()  # block Container type on CPD

            if use_flight and not self._obm:
                # Workaround for container connection type, we need to fetch COS details from space/project
                if self.type == DataConnectionTypes.CN:
                    # note: update flight parameters only if `connection_properties` was not set earlier
                    #       (e.x. by wml/autoi)
                    if not flight_parameters.get('connection_properties'):
                        flight_parameters = self._update_flight_parameters_with_connection_details(flight_parameters)

                data = self._download_data_from_flight_service(data_location=self,
                                                               binary=binary,
                                                               read_to_file=read_to_file,
                                                               flight_parameters=flight_parameters,
                                                               headers=headers,
                                                               enable_sampling=enable_sampling,
                                                               sampling_type=sampling_type,
                                                               number_of_batch_rows=number_of_batch_rows,
                                                               return_data_as_iterator=return_data_as_iterator,
                                                               _return_subsampling_stats=_return_subsampling_stats,
                                                               total_size_limit=total_size_limit,
                                                               total_nrows_limit=total_nrows_limit,
                                                               total_percentage_limit=total_percentage_limit)

            else:  # backward compatibility
                try:
                    with all_logging_disabled():
                        if self._check_if_connection_asset_is_s3():
                            cos_client = self._init_cos_client()
                            try:
                                if self._obm:
                                    data = self._download_obm_data_from_cos(cos_client=cos_client)

                                else:
                                    data = self._download_data_from_cos(cos_client=cos_client)

                            except Exception as cos_access_exception:
                                raise ConnectionError(
                                    f"Unable to access data object in cloud object storage with credentials supplied. "
                                    f"Error: {cos_access_exception}")
                        else:
                            data = self._download_data_from_nfs_connection()

                except Exception as e:
                    # do not try Flight is we are on the cloud
                    if self._wml_client is not None:
                        if not self._wml_client.ICP:
                            raise e

                    elif os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                        raise CannotReadSavedRemoteDataBeforeFit()

                    data = self._download_data_from_flight_service(data_location=self,
                                                                   binary=binary,
                                                                   read_to_file=read_to_file,
                                                                   flight_parameters=flight_parameters,
                                                                   headers=headers,
                                                                   enable_sampling=enable_sampling,
                                                                   sampling_type=sampling_type,
                                                                   number_of_batch_rows=number_of_batch_rows,
                                                                   _return_subsampling_stats=_return_subsampling_stats,
                                                                   total_size_limit=total_size_limit,
                                                                   total_nrows_limit=total_nrows_limit,
                                                                   total_percentage_limit=total_percentage_limit)

        if getattr(self._wml_client, '_internal', False):
            pass  # don't remove additional params if client is used internally
        else:
            # note: remove additional params and inline credentials added by _check_if_connection_asset_is_s3:
            [delattr(self.connection, attr) for attr in
             ['secret_access_key', 'access_key_id', 'endpoint_url', 'cos_type'] if
             hasattr(self.connection, attr)]
            # end note

        # create data statistics if data were not downloaded with flight:
        if not isinstance(data, tuple) and _return_subsampling_stats:
            data = (data, {"data_batch_size": sys.getsizeof(data),
                           "data_batch_nrows": len(data)})

        if binary:
            return data

        if raw or (self.auto_pipeline_params.get('prediction_column') is None
                   and self.auto_pipeline_params.get('prediction_columns') is None
                   and self.auto_pipeline_params.get('feature_columns') is None):
            return data

        else:
            if with_holdout_split:  # when this connection is training one
                if return_data_as_iterator:
                    raise WMLClientError("The flags `return_data_as_iterator` and `with_holdout_split` cannot be set both in the same time.")

                if _return_subsampling_stats:
                    X_train, X_holdout, y_train, y_holdout = self._recreate_holdout(data=data[0])
                    return X_train, X_holdout, y_train, y_holdout, data[1]
                else:
                    X_train, X_holdout, y_train, y_holdout = self._recreate_holdout(data=data)
                    return X_train, X_holdout, y_train, y_holdout

            else:  # when this data connection is a test / holdout one
                if return_data_as_iterator:
                    return data

                if _return_subsampling_stats:
                    if self.auto_pipeline_params.get('prediction_columns') or \
                            not self.auto_pipeline_params.get('prediction_column') or \
                            (self.auto_pipeline_params.get('prediction_column') and self.auto_pipeline_params.get(
                                'prediction_column') not in data[0].columns):
                        # timeseries dataset does not have prediction columns. Whole data set is returned:
                        test_X = data
                        return test_X
                    else:
                        test_X, test_y = self._recreate_holdout(data=data[0], with_holdout_split=False)
                        test_X[self.auto_pipeline_params.get('prediction_column', 'prediction_column')] = test_y
                        return test_X, data[1]

                else: # when this data connection is a test / holdout one and no subsampling stats are needed
                    if self.auto_pipeline_params.get('prediction_columns') or \
                            not self.auto_pipeline_params.get('prediction_column') or \
                            (self.auto_pipeline_params.get('prediction_column') and self.auto_pipeline_params.get(
                                'prediction_column') not in data.columns):
                        # timeseries dataset does not have prediction columns. Whole data set is returned:
                        test_X = data
                    else:
                        test_X, test_y = self._recreate_holdout(data=data, with_holdout_split=False)
                        test_X[self.auto_pipeline_params.get('prediction_column', 'prediction_column')] = test_y
                    return test_X  # return one dataframe

    def write(self, data: Union[str, 'DataFrame'], remote_name: str = None, **kwargs) -> None:
        """Upload file to a remote data storage.

        :param data: local path to the dataset or pandas.DataFrame with data
        :type data: str

        :param remote_name: name that dataset should be stored with in remote data storage
        :type remote_name: str
        """
        # enables flight automatically for CP4D 4.0.x, for 3.0 and 3.5 we do not have a flight service there
        use_flight = kwargs.get(
            'use_flight',
            bool((self._wml_client is not None or 'USER_ACCESS_TOKEN' in os.environ or 'RUNTIME_ENV_ACCESS_TOKEN_FILE' in os.environ) and (
                    self._wml_client.ICP_40 or self._wml_client.ICP_45 or self._wml_client.ICP_46)))

        flight_parameters = kwargs.get('flight_parameters', {})

        impersonate_header = kwargs.get('impersonate_header', None)

        headers = None
        if self._wml_client is None:
            token = self._get_token_from_environment()
            if token is None:
                raise ConnectionError("WML client missing. Please initialize WML client and pass it to "
                                      "DataConnection._wml_client property to be able to use this functionality.")

            else:
                headers = {'Authorization': f'Bearer {token}'}
        elif impersonate_header is not None:
            headers = self._wml_client._get_headers()
            headers['impersonate'] = impersonate_header

        # TODO: Remove S3 implementation
        if self.type == DataConnectionTypes.S3:
            raise ConnectionError("S3 DataConnection is deprecated! Please use data_asset_id instead.")

        elif self.type == DataConnectionTypes.CA or self.type == DataConnectionTypes.CN:
            if getattr(self._wml_client, 'ICP', False) and self.type == DataConnectionTypes.CN:
                raise ContainerTypeNotSupported()  # block Container type on CPD

            if self._check_if_connection_asset_is_s3():
                # do not try Flight if we are on the cloud
                if self._wml_client is not None and not self._wml_client.ICP and not use_flight:  # CLOUD
                    updated_remote_name = self._get_path_with_remote_name(self._to_dict(), remote_name)
                    cos_resource_client = self._init_cos_client()
                    if isinstance(data, str):
                        with open(data, "rb") as file_data:
                            cos_resource_client.Object(self.location.bucket, updated_remote_name).upload_fileobj(
                                Fileobj=file_data)

                    elif isinstance(data, DataFrame):
                        # note: we are saving csv in memory as a file and stream it to the COS
                        buffer = io.StringIO()
                        data.to_csv(buffer, index=False)
                        buffer.seek(0)

                        with buffer as f:
                            cos_resource_client.Object(self.location.bucket, updated_remote_name).upload_fileobj(
                                Fileobj=io.BytesIO(bytes(f.read().encode())))

                    else:
                        raise TypeError("data should be either of type \"str\" or \"pandas.DataFrame\"")
                # CP4D
                else:
                    # Workaround for container connection type, we need to fetch COS details from space/project
                    if self.type == DataConnectionTypes.CN:
                        # note: update flight parameters only if `connection_properties` was not set earlier
                        #       (e.x. by wml/autoi)
                        if not flight_parameters.get('connection_properties'):
                            flight_parameters = self._update_flight_parameters_with_connection_details(flight_parameters)

                    if isinstance(data, str):
                        self._upload_data_via_flight_service(file_path=data,
                                                             data_location=self,
                                                             remote_name=remote_name,
                                                             flight_parameters=flight_parameters,
                                                             headers=headers)

                    elif isinstance(data, DataFrame):
                        # note: we are saving csv in memory as a file and stream it to the COS
                        self._upload_data_via_flight_service(data=data,
                                                             data_location=self,
                                                             remote_name=remote_name,
                                                             flight_parameters=flight_parameters,
                                                             headers=headers)

                    else:
                        raise TypeError("data should be either of type \"str\" or \"pandas.DataFrame\"")

            else:
                if self._wml_client is not None and not self._wml_client.ICP and not use_flight:  # CLOUD
                    raise ConnectionError("Connections other than COS are not supported on a cloud yet.")
                # CP4D
                else:
                    if isinstance(data, str):
                        self._upload_data_via_flight_service(file_path=data,
                                                             data_location=self,
                                                             remote_name=remote_name,
                                                             flight_parameters=flight_parameters,
                                                             headers=headers)

                    elif isinstance(data, DataFrame):
                        # note: we are saving csv in memory as a file and stream it to the COS
                        self._upload_data_via_flight_service(data=data,
                                                             data_location=self,
                                                             remote_name=remote_name,
                                                             flight_parameters=flight_parameters,
                                                             headers=headers)

                    else:
                        raise TypeError("data should be either of type \"str\" or \"pandas.DataFrame\"")

            if getattr(self._wml_client, '_internal', False):
                pass  # don't remove additional params if client is used internally
            else:
                # note: remove additional params and inline credentials added by _check_if_connection_asset_is_s3:
                [delattr(self.connection, attr) for attr in ['secret_access_key', 'access_key_id', 'endpoint_url', 'cos_type'] if
                    hasattr(self.connection, attr)]
                # end note
        elif self.type == DataConnectionTypes.DS:
            if self._wml_client is not None and not self._wml_client.ICP and not use_flight:  # CLOUD
                raise ConnectionError("Write of data for Data Asset is not supported on Cloud.")

            elif self._wml_client is not None:
                if isinstance(data, str):
                    self._upload_data_via_flight_service(file_path=data,
                                                         data_location=self,
                                                         remote_name=remote_name,
                                                         flight_parameters=flight_parameters,
                                                         headers=headers)

                elif isinstance(data, DataFrame):
                    # note: we are saving csv in memory as a file and stream it to the COS
                    self._upload_data_via_flight_service(data=data,
                                                         data_location=self,
                                                         remote_name=remote_name,
                                                         flight_parameters=flight_parameters,
                                                         headers=headers)

                else:
                    raise TypeError("data should be either of type \"str\" or \"pandas.DataFrame\"")

            else:
                self._upload_data_via_flight_service(data=data,
                                                     data_location=self,
                                                     remote_name=remote_name,
                                                     flight_parameters=flight_parameters,
                                                     headers=headers)

    def _init_cos_client(self) -> 'resource':
        """Initiate COS client for further usage."""
        from ibm_botocore.client import Config
        if hasattr(self.connection, 'auth_endpoint') and hasattr(self.connection, 'api_key'):
            cos_client = resource(
                service_name='s3',
                ibm_api_key_id=self.connection.api_key,
                ibm_auth_endpoint=self.connection.auth_endpoint,
                config=Config(signature_version="oauth"),
                endpoint_url=self.connection.endpoint_url
            )

        else:
            cos_client = resource(
                service_name='s3',
                endpoint_url=self.connection.endpoint_url,
                aws_access_key_id=self.connection.access_key_id,
                aws_secret_access_key=self.connection.secret_access_key
            )
        return cos_client

    def _validate_cos_resource(self):
        cos_client = self._init_cos_client()
        try:
            files = cos_client.Bucket(self.location.bucket).objects.all()
            next(x for x in files if x.key == self.location.path)
        except Exception as e:
            raise NotExistingCOSResource(self.location.bucket, self.location.path)

    def _update_flight_parameters_with_connection_details(self, flight_parameters):
        with all_logging_disabled():
            self._check_if_connection_asset_is_s3()
            connection_properties = {
                "bucket": self.location.bucket,
                "url": self.connection.endpoint_url
            }
            if hasattr(self.connection, 'auth_endpoint') and hasattr(self.connection, 'api_key'):
                connection_properties["iam_url"] = self.connection.auth_endpoint
                connection_properties["api_key"] = self.connection.api_key
                connection_properties["resource_instance_id"] = self.connection.resource_instance_id

            else:
                connection_properties["secret_key"] = self.connection.secret_access_key
                connection_properties["access_key"] = self.connection.access_key_id

            flight_parameters.update({"connection_properties": connection_properties})
            flight_parameters.update({"datasource_type": {"entity": {"name": self._datasource_type}}})

        return flight_parameters


# TODO: Remove S3 Implementation for connection
class S3Connection(BaseConnection):
    """Connection class to COS data storage in S3 format.

    :param endpoint_url: S3 data storage url (COS)
    :type endpoint_url: str

    :param access_key_id: access_key_id of the S3 connection (COS)
    :type access_key_id: str, optional

    :param secret_access_key: secret_access_key of the S3 connection (COS)
    :type secret_access_key: str, optional

    :param api_key: API key of the S3 connection (COS)
    :type api_key: str, optional

    :param service_name: service name of the S3 connection (COS)
    :type service_name: str, optional

    :param auth_endpoint: authentication endpoint url of the S3 connection (COS)
    :type auth_endpoint: str, optional
    """

    def __init__(self, endpoint_url: str, access_key_id: str = None, secret_access_key: str = None,
                 api_key: str = None, service_name: str = None, auth_endpoint: str = None,
                 resource_instance_id: str = None, _internal_use=False) -> None:

        if not _internal_use:
            warn(message="S3 DataConnection is deprecated! Please use data_asset_id instead.")

        if (access_key_id is None or secret_access_key is None) and (api_key is None or auth_endpoint is None):
            raise InvalidCOSCredentials(reason='You need to specify (access_key_id and secret_access_key) or'
                                               '(api_key and auth_endpoint)')

        if secret_access_key is not None:
            self.secret_access_key = secret_access_key

        if api_key is not None:
            self.api_key = api_key

        if service_name is not None:
            self.service_name = service_name

        if auth_endpoint is not None:
            self.auth_endpoint = auth_endpoint

        if access_key_id is not None:
            self.access_key_id = access_key_id

        if endpoint_url is not None:
            self.endpoint_url = endpoint_url

        if resource_instance_id is not None:
            self.resource_instance_id = resource_instance_id


class S3Location(BaseLocation):
    """Connection class to COS data storage in S3 format.

    :param bucket: COS bucket name
    :type bucket: str

    :param path: COS data path in the bucket
    :type path: str

    :param excel_sheet: name of excel sheet if pointed dataset is excel file used for Batched Deployment scoring
    :type excel_sheet: str, optional

    :param model_location: path to the pipeline model in the COS
    :type model_location: str, optional

    :param training_status: path to the training status json in COS
    :type training_status: str, optional
    """

    def __init__(self, bucket: str, path: str, **kwargs) -> None:
        self.bucket = bucket
        self.path = path

        if kwargs.get('model_location') is not None:
            self._model_location = kwargs['model_location']

        if kwargs.get('training_status') is not None:
            self._training_status = kwargs['training_status']

        if kwargs.get('excel_sheet') is not None:
            self.sheet_name = kwargs['excel_sheet']
            self.file_format = "xls"

    def _get_file_size(self, cos_resource_client: 'resource') -> 'int':
        try:
            size = cos_resource_client.Object(self.bucket, self.path).content_length
        except ClientError:
            size = 0
        return size

    def get_location(self) -> str:
        if hasattr(self, "file_name"):
            return self.file_name
        else:
            return self.path


class ContainerLocation(BaseLocation):
    """Connection class to default COS in user Project/Space."""

    def __init__(self, path: Optional[str] = None, **kwargs) -> None:
        if path is None:
            self.path = "default_autoai_out"

        else:
            self.path = path

        self.bucket = None

        if kwargs.get('model_location') is not None:
            self._model_location = kwargs['model_location']

        if kwargs.get('training_status') is not None:
            self._training_status = kwargs['training_status']

    def to_dict(self) -> dict:
        _dict = super().to_dict()

        if 'bucket' in _dict and _dict['bucket'] is None:
            del _dict['bucket']

        return _dict

    @classmethod
    def _set_path(cls, path: str) -> 'ContainerLocation':
        location = cls()
        location.path = path
        return location

    def _get_file_size(self):
        pass


class FSLocation(BaseLocation):
    """Connection class to File Storage in CP4D."""

    def __init__(self, path: Optional[str] = None) -> None:
        if path is None:
            self.path = "/{option}/{id}" + f"/assets/auto_ml/auto_ml.{uuid.uuid4()}/wml_data"

        else:
            self.path = path

    @classmethod
    def _set_path(cls, path: str) -> 'FSLocation':
        location = cls()
        location.path = path
        return location

    def _save_file_as_data_asset(self, workspace: 'WorkSpace') -> 'str':

        asset_name = self.path.split('/')[-1]
        if self.path:
            data_asset_details = workspace.wml_client.data_assets.create(asset_name, self.path)
            return workspace.wml_client.data_assets.get_uid(data_asset_details)
        else:
            raise MissingValue('path', reason="Incorrect initialization of class FSLocation")

    def _get_file_size(self, workspace: 'WorkSpace') -> 'int':
        # note if path is not file then returned size is 0
        try:
            # note: try to get file size from remote server
            url = workspace.wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() \
                  + f"/{self.path.split('/assets/')[-1]}"
            path_info_response = requests.head(url, headers=workspace.wml_client._get_headers(),
                                               params=workspace.wml_client._params())
            if path_info_response.status_code != 200:
                raise ApiRequestFailure(u"Failure during getting path details", path_info_response)
            path_info = path_info_response.headers
            if 'X-Asset-Files-Type' in path_info and path_info['X-Asset-Files-Type'] == 'file':
                size = path_info['X-Asset-Files-Size']
            else:
                size = 0
            # -- end note
        except (ApiRequestFailure, AttributeError):
            # note try get size of file from local fs
            size = os.stat(path=self.path).st_size if os.path.isfile(path=self.path) else 0
            # -- end note
        return size


class AssetLocation(BaseLocation):

    def __init__(self, asset_id: str) -> None:
        self._wsd = self._is_wsd()
        self.href = None
        self._initial_asset_id = asset_id
        self.__wml_client = None

        if self._wsd:
            self._asset_name = None
            self._asset_id = None
            self._local_asset_path = None
        else:
            self.id = asset_id

    def _get_bucket(self, client) -> str:
        """Try to get bucket from data asset."""
        connection_id = self._get_connection_id(client)
        conn_details = client.connections.get_details(connection_id)
        bucket = conn_details.get('entity', {}).get('properties', {}).get('bucket')

        if bucket is None:
            asset_details = client.data_assets.get_details(self.id)
            connection_path = asset_details['entity'].get('folder_asset', {}).get('connection_path')
            if connection_path is None:
                attachment_content = self._get_attachment_details(client)
                connection_path = attachment_content.get('connection_path')

            bucket = connection_path.split('/')[1]

        return bucket

    def _get_attachment_details(self, client) -> dict:
        if self.id is None and self.href:
            items = self.href.split('/')
            self.id = items[-1].split('?')[0]

        asset_details = client.data_assets.get_details(self.id)

        if 'attachment_id' in asset_details.get('metadata'):
            attachment_id = asset_details['metadata']['attachment_id']

        else:
            attachment_id = asset_details['attachments'][0]['id']

        attachment_url = client.service_instance._href_definitions.get_data_asset_href(self.id)
        attachment_url = f"{attachment_url}/attachments/{attachment_id}"

        if client.ICP:
            attachment = requests.get(attachment_url, headers=client._get_headers(),
                                      params=client._params())

        else:
            attachment = requests.get(attachment_url, headers=client._get_headers(),
                                      params=client._params())

        if attachment.status_code != 200:
            raise ApiRequestFailure(u"Failure during getting attachment details", attachment)

        return attachment.json()

    def _get_connection_id(self, client) -> str:
        attachment_content = self._get_attachment_details(client)

        return attachment_content.get('connection_id')

    @classmethod
    def _is_wsd(cls):
        if os.environ.get('USER_ACCESS_TOKEN') or os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE'):
            return False

        try:
            from project_lib import Project
            try:
                with all_logging_disabled():
                    access = Project.access()
                return True
            except RuntimeError:
                pass
        except ModuleNotFoundError:
            pass

        return False

    @classmethod
    def _set_path(cls, href: str) -> 'AssetLocation':
        items = href.split('/')
        _id = items[-1].split('?')[0]
        location = cls(_id)
        location.href = href
        return location

    def _get_file_size(self, workspace: 'WorkSpace', *args) -> 'int':
        if self._wsd:
            return self._wsd_get_file_size()
        else:
            asset_info_response = requests.get(
                workspace.wml_client.service_instance._href_definitions.get_data_asset_href(self.id),
                params=workspace.wml_client._params(),
                headers=workspace.wml_client._get_headers())
            if asset_info_response.status_code != 200:
                raise ApiRequestFailure(u"Failure during getting asset details", asset_info_response)
            return asset_info_response.json()['metadata'].get('size')

    def _wsd_setup_local_asset_details(self) -> None:
        if not self._wsd:
            raise NotWSDEnvironment()

        # note: set local asset file from asset_id
        project = get_project()
        project_id = project.get_metadata()["metadata"]["guid"]

        local_assets = project.get_files()

        # note: reuse local asset_id when object is reused more times
        if self._asset_id is None:
            local_asset_id = self._initial_asset_id

        else:
            local_asset_id = self._asset_id
        # --- end note

        if local_asset_id not in str(local_assets):
            raise MissingLocalAsset(local_asset_id, reason="Provided asset_id cannot be found on WS Desktop.")

        else:
            for asset in local_assets:
                if asset['asset_id'] == local_asset_id:
                    asset_name = asset['name']
                    self._asset_name = asset_name
                    self._asset_id = local_asset_id

            local_asset_path = f"{os.path.abspath('.')}/{project_id}/assets/data_asset/{asset_name}"
            self._local_asset_path = local_asset_path

    def _wsd_move_asset_to_server(self, workspace: 'WorkSpace') -> None:
        if not self._wsd:
            raise NotWSDEnvironment()

        if not self._local_asset_path or self._asset_name or self._asset_id:
            self._wsd_setup_local_asset_details()

        remote_asset_details = workspace.wml_client.data_assets.create(self._asset_name, self._local_asset_path)
        self.href = remote_asset_details['metadata']['href']

    def _wsd_get_file_size(self) -> 'int':
        if not self._wsd:
            raise NotWSDEnvironment()

        if not self._local_asset_path or self._asset_name or self._asset_id:
            self._wsd_setup_local_asset_details()
        return os.stat(path=self._local_asset_path).st_size if os.path.isfile(path=self._local_asset_path) else 0

    @classmethod
    def list_wsd_assets(cls):
        if not cls._is_wsd():
            raise NotWSDEnvironment

        project = get_project()
        return project.get_files()

    def to_dict(self) -> dict:
        """Return a json dictionary representing this model."""
        _dict = vars(self).copy()

        if _dict.get('id', False) is None and _dict.get('href'):
            items = self.href.split('/')
            _dict['id'] = items[-1].split('?')[0]

        del _dict['_wsd']
        del _dict[f"_{self.__class__.__name__}__wml_client"]

        if self._wsd:
            del _dict['_asset_name']
            del _dict['_asset_id']
            del _dict['_local_asset_path']

        del _dict['_initial_asset_id']

        return _dict

    @property
    def wml_client(self):
        return self.__wml_client

    @wml_client.setter
    def wml_client(self, var):
        self.__wml_client = var

        if self.__wml_client:
            self.href = self.__wml_client.service_instance._href_definitions.get_base_asset_href(self._initial_asset_id)
        else:
            self.href = f'/v2/assets/{self._initial_asset_id}'

        if not self._wsd:
            if self.__wml_client:
                if self.__wml_client.default_space_id:
                    self.href = f'{self.href}?space_id={self.__wml_client.default_space_id}'
                else:
                    self.href = f'{self.href}?project_id={self.__wml_client.default_project_id}'


class ConnectionAssetLocation(BaseLocation):
    """Connection class to COS data storage.

    :param bucket: COS bucket name
    :type bucket: str

    :param file_name: COS data path in the bucket
    :type file_name: str

    :param model_location: path to the pipeline model in the COS
    :type model_location: str, optional

    :param training_status: path to the training status json in COS
    :type training_status: str, optional
    """

    def __init__(self, bucket: str, file_name: str, **kwargs) -> None:
        self.bucket = bucket
        self.file_name = file_name
        self.path = file_name

        if kwargs.get('model_location') is not None:
            self._model_location = kwargs['model_location']

        if kwargs.get('training_status') is not None:
            self._training_status = kwargs['training_status']

    def _get_file_size(self, cos_resource_client: 'resource') -> 'int':
        try:
            size = cos_resource_client.Object(self.bucket, self.path).content_length
        except ClientError:
            size = 0
        return size

    def to_dict(self) -> dict:
        """Return a json dictionary representing this model."""
        return vars(self)


class ConnectionAsset(BaseConnection):
    """Connection class for Connection Asset.

    :param connection_id: connection asset ID
    :type connection_id: str
    """

    def __init__(self, connection_id: str):
        self.id = connection_id


class NFSConnection(BaseConnection):
    """Connection class to file storage in CP4D of NFS format.

    :param asset_id: asset ID from the project on CP4D
    :type asset_id: str
    """

    def __init__(self, asset_id: str):
        self.asset_id = asset_id
        self.id = asset_id


class NFSLocation(BaseLocation):
    """Location class to file storage in CP4D of NFS format.

    :param path: data path form the project on CP4D
    :type path: str
    """

    def __init__(self, path: str):
        self.path = path
        self.id = None
        self.file_name = None

    def _get_file_size(self, workspace: 'Workspace', *args) -> 'int':
        params = workspace.wml_client._params().copy()
        params['path'] = self.path
        params['detail'] = 'true'

        href = workspace.wml_client.connections._href_definitions.get_connection_by_id_href(self.id) + '/assets'
        asset_info_response = requests.get(href,
                                           params=params, headers=workspace.wml_client._get_headers(None))
        if asset_info_response.status_code != 200:
            raise Exception(u"Failure during getting asset details", asset_info_response.json())
        return asset_info_response.json()['details']['file_size']

    def get_location(self) -> str:
        if hasattr(self, "file_name"):
            return self.file_name
        else:
            return self.path


class CP4DAssetLocation(AssetLocation):
    """Connection class to data assets in CP4D.

    :param asset_id: asset ID from the project on CP4D
    :type asset_id: str
    """

    def __init__(self, asset_id: str) -> None:
        super().__init__(asset_id)
        warning_msg = ("Depreciation Warning: Class CP4DAssetLocation is no longer supported and will be removed."
                       "Use AssetLocation instead.")
        print(warning_msg)

    def _get_file_size(self, workspace: 'WorkSpace', *args) -> 'int':
        return super()._get_file_size(workspace)


class WMLSAssetLocation(AssetLocation):
    """Connection class to data assets in WML Server.

    :param asset_id: asset ID of the file loaded on space in WML Server
    :type asset_id: str
    """

    def __init__(self, asset_id: str) -> None:
        super().__init__(asset_id)
        warning_msg = ("Depreciation Warning: Class WMLSAssetLocation is no longer supported and will be removed."
                       "Use AssetLocation instead.")
        print(warning_msg)

    def _get_file_size(self, workspace: 'WorkSpace', *args) -> 'int':
        return super()._get_file_size(workspace)


class CloudAssetLocation(AssetLocation):
    """Connection class to data assets as input data references to batch deployment job on Cloud.

    :param asset_id: asset ID of the file loaded on space on Cloud
    :type asset_id: str
    """

    def __init__(self, asset_id: str) -> None:
        super().__init__(asset_id)
        self.href = self.href
        warning_msg = ("Depreciation Warning: Class CloudAssetLocation is no longer supported and will be removed."
                       "Use AssetLocation instead.")
        print(warning_msg)

    def _get_file_size(self, workspace: 'WorkSpace', *args) -> 'int':
        return super()._get_file_size(workspace)


class WSDAssetLocation(BaseLocation):
    """Connection class to data assets in WS Desktop.

    :param asset_id: asset ID from the project on WS Desktop
    :type asset_id: str
    """

    def __init__(self, asset_id: str) -> None:
        self.href = None
        self._asset_name = None
        self._asset_id = None
        self._local_asset_path = None
        self._initial_asset_id = asset_id
        self.__wml_client = None

        warning_msg = ("Depreciation Warning: Class WSDAssetLocation is no longer supported and will be removed."
                       "Use AssetLocation instead.")
        print(warning_msg)

    @classmethod
    def list_assets(cls):
        project = get_project()
        return project.get_files()

    def _setup_local_asset_details(self) -> None:
        # note: set local asset file from asset_id
        project = get_project()
        project_id = project.get_metadata()["metadata"]["guid"]

        local_assets = project.get_files()

        # note: reuse local asset_id when object is reused more times
        if self._asset_id is None:
            local_asset_id = self.href.split('/')[3].split('?space_id')[0]

        else:
            local_asset_id = self._asset_id
        # --- end note

        if local_asset_id not in str(local_assets):
            raise MissingLocalAsset(local_asset_id, reason="Provided asset_id cannot be found on WS Desktop.")

        else:
            for asset in local_assets:
                if asset['asset_id'] == local_asset_id:
                    asset_name = asset['name']
                    self._asset_name = asset_name
                    self._asset_id = local_asset_id

            local_asset_path = f"{os.path.abspath('.')}/{project_id}/assets/data_asset/{asset_name}"
            self._local_asset_path = local_asset_path

    def _move_asset_to_server(self, workspace: 'WorkSpace') -> None:
        if not self._local_asset_path or self._asset_name or self._asset_id:
            self._setup_local_asset_details()

        remote_asset_details = workspace.wml_client.data_assets.create(self._asset_name, self._local_asset_path)
        self.href = remote_asset_details['metadata']['href']

    @classmethod
    def _set_path(cls, href: str) -> 'WSDAssetLocation':
        location = cls('.')
        location.href = href
        return location

    @property
    def wml_client(self):
        return self.__wml_client

    @wml_client.setter
    def wml_client(self, var):
        self.__wml_client = var

        if self.__wml_client:
            self.href = self.__wml_client.service_instance._href_definitions.get_base_asset_href(self._initial_asset_id)
        else:
            self.href = f'/v2/assets/{self._initial_asset_id}'

    def to_dict(self) -> dict:
        """Return a json dictionary representing this model."""
        _dict = vars(self).copy()
        del _dict['_asset_name']
        del _dict['_asset_id']
        del _dict['_local_asset_path']
        del _dict[f"_{self.__class__.__name__}__wml_client"]
        del _dict['_initial_asset_id']

        return _dict

    def _get_file_size(self) -> 'int':
        if not self._local_asset_path or self._asset_name or self._asset_id:
            self._setup_local_asset_details()
        return os.stat(path=self._local_asset_path).st_size if os.path.isfile(path=self._local_asset_path) else 0


class DeploymentOutputAssetLocation(BaseLocation):
    """Connection class to data assets where output of batch deployment will be stored.

    :param name: name of .csv file which will be saved as data asset
    :type name: str
    :param description: description of the data asset
    :type description: str, optional
    """

    def __init__(self, name: str, description: str = "") -> None:
        self.name = name
        self.description = description


class DatabaseLocation(BaseLocation):
    """Location class to Database.

    :param schema_name: database schema name
    :type schema_name: str
    :param table_name: database table name
    :type table_name: str
    """

    def __init__(self, schema_name: str, table_name: str, **kwargs) -> None:
        self.schema_name = schema_name
        self.table_name = table_name

    def _get_file_size(self) -> None:
        raise NotImplementedError()



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import logging
import numpy as np

from ibmfl.party.training.local_training_handler import \
    LocalTrainingHandler

logger = logging.getLogger(__name__)


class PFNMLocalTrainingHandler(LocalTrainingHandler):

    def train(self,  fit_params=None):
        """
        Train locally using fl_model. At the end of training, a
        model_update with the new model information is generated and
        send through the connection.

        :param fit_params: (optional) Query instruction from aggregator
        :type fit_params: `dict`
        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """

        train_data, (_) = self.data_handler.get_data()
        y = train_data[1]

        y_unq, y_cnts = np.unique(y, return_counts=True)
        _class_counts = {
            int(cls_label): int(cls_counts) for (cls_label, cls_counts) in
            zip(y_unq, y_cnts)
        }

        self.update_model(fit_params['model_update'])

        logger.info('Local training started...')

        self.fl_model.fit_model(train_data, fit_params)

        update = self.fl_model.get_model_update()
        update.add('class_counts', _class_counts)

        # Different frameworks have different return dimension.
        # PyTorch returns (features_out, features_in) dimensional weights for linear layers
        # Keras returns (features_in, features_out) dimensional weights for linear layers
        # Thus, this flag indicates if a transpose of the weight matrix is required or not

        update.add('transpose_weight', self.fl_model.model_type == 'PyTorch')
        logger.info('Local training done, generating model update...')

        return update



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""Connection class which uses flask and request libraries to create a server
client combo
"""
import logging
import socket

import requests
import asyncio
import signal
import random
import websockets
import pathlib
from websockets import exceptions
import string
import argparse
import threading
import time
import ssl
from _collections import deque
import sys
import traceback

from ibmfl.connection.connection import ConnectionStatus
from ibmfl.connection.connection import FLConnection, FLSender, FLReceiver
from ibmfl.message.message_type import MessageType
from ibmfl.message.message import Message
from ibmfl.message.serializer_types import SerializerTypes
from ibmfl.message.serializer_factory import SerializerFactory

logger = logging.getLogger(__name__)


# Maintained only on the party side
# Used to signal to websockets that there is a message ready to be sent to the aggregator
pSendEvt = threading.Event()

# Used to signal that there is a message that was received from the aggregator
pRecvEvt = threading.Event()



# Used to store messages to be sent to aggregator
party_sbuffer = deque([])
# Used to store messages received from aggregator
party_rbuffer = deque([])

pRouter = None


class WSConnection(FLConnection):

    def __init__(self, config):
        """
        Initializes the connection object
        :param config: dictionary of configuration provided to connection
        :type config: `dict`
        """

        pSendEvt.clear()
        pRecvEvt.clear()
        # comes from the config file
        self.started = False
        self.stopped = False
        self.settings = config
        self.receiver = None
        self.sender = None
        self.loop1 = None
        self.loop2 = None
        self.stop1 = None
        self.stop2 = None
        self.mainLoop = None
        self.flagForOutstandingMessageResponseToBeReceived = False
        # identity/name of the party

        if 'id' in config:
            self.party = config['id']
        else:
            # TODO: Fix error if no rts id
            self.party = 'DEFAULT'

        # Send Port is used for messages initiated from aggregator
        if 'wssendport' in config:
            self.wssendport = config['wssendport']
        else:
            # set a default value to be aggregator
            self.wssendport = "/aggregator"
        # Recv port is used for messages initiated from party
        if 'wsrecvport' in config:
            self.wsrecvport = config['wsrecvport']
        else:
            # set a default value to be aggregator
            self.wsrecvport = "/aggregator1"


    async def PartySendLoop(self):
        """
        This is started at the party.
        It reads from the party_sbuffer, and synchronously sends each message to the aggregator
        and waits for a response from the aggregator.
        SendLoop is a misnomer, like in Flask Connection.
        It merely means that messages are initiated by the party.
        """

        uri = "wss://" + self.aggInfo['ip'] + str(self.wsrecvport)
        logger.debug("PartySendLoop started = " + uri)
        ssl_context_party = True
        if self.ssl_context is not None:
            ssl_context_party = ssl.SSLContext()

        while not self.stopped:
            if len(party_sbuffer) == 0:
                logger.info("PartySendLoop: Holding for message to send" )
                pSendEvt.wait()
            try:
                logger.info("PartySendLoop: Number of active messages ready to send: " + str(len(party_sbuffer)))
                msg = party_sbuffer.popleft()
            except IndexError:
                logger.debug("No message to send")
                continue
            pSendEvt.clear()
            logger.debug("PartySendLoop: Sending message to aggregator")
            headers = None
            if self.authToken != None:
                headers = { 'Authorization': self.authToken, 'rtsid': self.party }
            try:
                async with websockets.connect(uri, max_size=2 ** 29, read_limit=2 ** 29, write_limit=2**29, ssl=ssl_context_party, extra_headers=headers) as websocket:

                    if self.flagForOutstandingMessageResponseToBeReceived == False:
                        logger.debug(
                            "PartySendLoop: Attempting to send message to aggregator")
                        try:
                            await websocket.send(msg)

                        except Exception as ex:
                            logger.error("PartySendLoop exception when attempting to send message: " + str(ex))
                            party_sbuffer.appendleft(msg)
                            raise
                        logger.debug(
                            "PartySendLoop: Message sent successfully; awaiting message response")

                    self.flagForOutstandingMessageResponseToBeReceived = True
                    resp = await websocket.recv()
                    self.flagForOutstandingMessageResponseToBeReceived = False
                    party_rbuffer.append(resp)
                    pRecvEvt.set()

            except websockets.exceptions.ConnectionClosedError as ex:
                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed abnormally. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.ConnectionClosedOK as ex:
                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed unexpectedly. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.WebSocketException as ex:
                logger.error(
                    "PartySendLoop : WebsocketException. Details: " + str(ex))
                return False
            except socket.gaierror as ex:
                logger.error("PartySendLoop : Error connecting to " + uri + ". Exception is " + str(ex))
                return False
            except Exception as ex:
                logger.error("PartySendLoop : Exception is " + str(ex))
                return False
        logger.debug("PartySendLoop ended ")
        return True

    async def PartyRecvLoop(self):
        """
        This is started at the party.
        It establishes a persistent connection to the aggregator by sending the name of the party
        The aggregator can then send messages to the party synchronously
        to which the party replies
        """
        pName = self.party
        uri = "wss://" + self.aggInfo['ip'] + str(self.wssendport)
        logger.debug("PartyRecvLoop started = " + uri)

        ssl_context_party2 = True
        if self.ssl_context is not None:
            ssl_context_party2 = ssl.SSLContext()

        headers = None
        if self.authToken != None:
            headers = { 'Authorization': self.authToken }
        while not self.stopped:
            try:
                async with websockets.connect(uri, close_timeout=100, max_size=2 ** 29, read_limit=2 ** 29, write_limit=2**29, ssl=ssl_context_party2, extra_headers=headers) as websocket:
                    await websocket.send(pName)
                    async for rmess in websocket:
                        if rmess == "HEARTBEAT":
                            logger.info("Received Heartbeat from Aggregator")
                            continue
                        serializer = SerializerFactory(
                            SerializerTypes.JSON_PICKLE).build()
                        recv_msg = serializer.deserialize(rmess)
                        logger.debug(
                            "PartyRecvLoop : Received " + str(recv_msg))
                        request_path = str(recv_msg.message_type)
                        logger.debug(
                            "REQUEST PATH IN PARTY IS " + str(request_path))

                        # check if error
                        if recv_msg.message_type == MessageType.ERROR_AUTH.value :
                            logger.info(
                            "PartyRecvLoop : Auth Error. Please verify training ID, RTS, project permission and token")
                            self.stop()
                            break;

                        handler, kwargs = self.router.get_handler(
                            request_path=request_path)

                        if handler is None:
                            logger.error(
                                'PartyRecvLoop : Invalid Request ! Routing it to default handler')
                            handler, kwargs = self.router.get_handler(
                                request_path='default')
                        try:
                            res_message = handler(recv_msg)
                        except Exception as ex:
                            res_message = Message()
                            data = {'status': 'error', 'message': str(ex)}
                            res_message.set_data(data)

                        response = serializer.serialize(res_message)
                        await websocket.send(response)
                        logger.info("PartyRecvLoop : sent response")

                        if recv_msg.message_type == MessageType.STOP.value :
                            logger.info(
                            "PartyRecvLoop : Received STOP message" )
                            self.stop()
                            break;

            except websockets.exceptions.ConnectionClosedError as ex:
                if self.stopped != True:
                    logger.error(
                        "PartyRecvLoop : Connection closed abnormally. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.ConnectionClosedOK as ex:
                if self.stopped != True:
                    logger.error(
                        "PartyRecvLoop : Connection closed unexpectedly. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.WebSocketException as ex:
                logger.error(
                    "PartyRecvLoop : WebsocketException. Details: " + str(ex))
                return False
            except socket.gaierror as ex:
                logger.error("PartyRecvLoop : Error connecting to " + uri + ". Exception is " + str(ex))
                return False
            except Exception as ex:
                logger.error("PartyRecvLoop : Exception is " + str(ex))
                return False
        logger.debug("PartyRecvLoop ended ")
        return True

    def PartySendLoopThread(self, loop):
        logger.info("**** PartySendLoopThread")
        # Need to host the loop in its own thread for asyncio to work properly
        pslt_result = loop.run_until_complete(self.PartySendLoop())
        logger.debug("**** PartySendLoopThread RESULT = " + str(pslt_result))
        if pslt_result == False:
            #if error when sending message, then unblock mainThread so it can fail.
            pRecvEvt.set()
            self.stop()

    def PartyRecvLoopThread(self, loop):
        logger.info("**** PartyRecvLoopThread")
        # Need to host the loop in its own thread for asyncio to work properly
        prlt_result = loop.run_until_complete(self.PartyRecvLoop())
        logger.debug("**** PartyRecvLoopThread RESULT = " + str(prlt_result))
        if prlt_result == False:
            self.stop()


    def initialize(self, router, aggInfo=None, ssl_context=None, authToken=None):
        """Initialize receiver and sender """
        self.router = router
        self.ssl_context = ssl_context

        if aggInfo == None:
            logger.error("Aggregator information not specified. Exiting.")
            exit()
        logger.info("WSConnection : Initialize Party Communications")
        self.party = self.settings['id']
        self.aggInfo = aggInfo
        self.authToken = authToken

        # Each websocket connection needs a separate asyncio event loop
        self.loop1 = asyncio.new_event_loop()
        t1 = threading.Thread(
            target=self.PartySendLoopThread, args=(self.loop1,), daemon=True)
        t1.start()

        # Each websocket connection needs a separate asyncio event loop
        self.loop2 = asyncio.new_event_loop()
        t2 = threading.Thread(
            target=self.PartyRecvLoopThread, args=(self.loop2,), daemon=True)
        t2.start()

    def initialize_receiver(self, router=None):
        """Basically does nothing
        :param router: Router object describing the routes for each request
            which are passed down to PH
        :type router: `Router`
        """
        # print(router)
        self.router = router
        pRouter = router
        self.receiver = WSReceiver(router, self.settings, self)
        self.receiver.connection = self
        logger.debug('Receiver Initialized')
        self.receiver.initialize()
        self.status = ConnectionStatus.INITIALIZED

    def initialize_sender(self):
        """BAsically does nothing
        """
        self.sender = WSSender(self.settings)
        self.sender.initialize()
        self.status = ConnectionStatus.INITIALIZED

    def start(self):
        """
        Set self.started to True
        """
        # self.start_receiver()
        self.started = True
        self.status = ConnectionStatus.STARTED
        self.SENDER_STATUS = ConnectionStatus.STARTED

    def stop(self):
        """
        set self.stopped to true
        """
        self.stopped = True
        logger.debug('Stopping Receiver and Sender')
        pSendEvt.set()
        self.status = ConnectionStatus.STOPPED
        try:
            if self.stop1 is not None:
                self.stop1.cancel()
            if self.stop2 is not None:
                self.stop2.set_result(None)
            if self.mainLoop is not None:
                self.mainLoop.call_soon_threadsafe(self.mainLoop.stop)
        except Exception as ex:
            logger.error("Exception when stopping connection: " + str(ex))

    def get_connection_config(self):
        pass

class WSReceiver(FLReceiver):

    def __init__(self, router, settings, connection):
        """
        Does nothing
        """

        self.router = router
        self.settings = settings
        self.connection = connection

    def shutdown_server(self):
        pass

    def initialize(self):
        """
        Does nothing
        """
        logger.debug('Initializing WSReceiver')


    def receive_message(self, party):
        logger.error('OOOOO We are in trouble. This should not be called')

    def start(self):
        pass

    def stop(self):
        pass


class WSSender(FLSender):

    """
    Basic implementation using the request package of python to send
    requests to Rest endpoints.
    """

    def __init__(self, settings):
        """
        Does nothing
        """
        self.settings = settings
        logger.debug("WS Sender Init. My info is " + str(settings))

    def initialize(self):
        """Does all the setups required for the sender. This method should also
        check for availability of resources, open ports, certificates etc.,
        if required
        """
        logger.info('Websockets Sender initialized')

    def send_message(self, destination, message):
        """
        used for sending all the requests. Message object should be
        validated and endpoint should be decided based on message codes.
        :param destination: information about the destination to which message
        should be forwarded
        :type destination: `dict`
        :param message: message object constructed by aggregator/party
        :type message: `Message`
        :return : response object
        :rtype : `Response`
        """
        if 'id' in self.settings:
            message.add_sender_info(self.settings['id'])
        else:
            # set a default value to be Aggregator
            message.add_sender_info('Aggregator')

        serializer = SerializerFactory(SerializerTypes.JSON_PICKLE).build()
        message = serializer.serialize(message)

        logger.info("Sending serialized message to aggregator")
        party_sbuffer.append(message)
        pSendEvt.set()
        pRecvEvt.wait()
        response = None
        try:
            response = party_rbuffer.popleft()
        except IndexError:
            logger.debug("[WSSender] No message response received, after attempt to send message")

        pRecvEvt.clear()
        response_message = None
        if response is not None:
            # inspect response
            if isinstance(response, str):
                logger.info('Received String message as response: ' + response)
                response_message = Message(
                    MessageType.REGISTER.value, data={'status': 'success', 'response': response})
            else:
                # deserialize response
                response_message = serializer.deserialize(response)
                logger.info(
                    'Received serialized message as response: ' + str(response_message))

        return response_message

    def cleanup(self):
        """
        Cleanup the Sender object and close the hanging connections if any
        """
        logger.info('Cleaning up Websockets Client')



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.metanames import ExportMetaNames
import os
import json
from ibm_watson_machine_learning.wml_client_error import WMLClientError, UnexpectedType, ApiRequestFailure
from ibm_watson_machine_learning.wml_resource import WMLResource
_DEFAULT_LIST_LENGTH = 50

class Export(WMLResource):
    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)

        self._client = client
        self.ConfigurationMetaNames = ExportMetaNames()

    def start(self, meta_props, space_id=None, project_id=None):
        """Start the export. Either space_id or project_id has to be provided and is mandatory.
        ALL_ASSETS is by default False. No need to provide explicitly unless it has to be set to True.
        Either ALL_ASSETS or ASSET_TYPES or ASSET_IDS has to be given in the meta_props. Only one of these can be
        provided.

        In the `meta_props`:

        ALL_ASSETS is a boolean. When set to True, it exports all assets in the given space.
        ASSET_IDS is an array containing the list of assets ids to be exported.
        ASSET_TYPES is for providing the asset types to be exported. All assets of that asset type will be exported.

                Eg: wml_model, wml_model_definition, wml_pipeline, wml_function, wml_experiment,
                software_specification, hardware_specification, package_extension, script

        :param meta_props: meta data,
            to see available meta names use ``client.export_assets.ConfigurationMetaNames.get()``
        :type meta_props: dict
        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project: str, optional

        :return: Response json
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                client.export_assets.ConfigurationMetaNames.ASSET_IDS: ["13a53931-a8c0-4c2f-8319-c793155e7517",
                                                                        "13a53931-a8c0-4c2f-8319-c793155e7518"]}

            details = client.export_assets.start(meta_props=metadata, space_id="98a53931-a8c0-4c2f-8319-c793155e4598")

        .. code-block:: python

            metadata = {
                client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                client.export_assets.ConfigurationMetaNames.ASSET_TYPES: ["wml_model"]}

            details = client.export_assets.start(meta_props=metadata, space_id="98a53931-a8c0-4c2f-8319-c793155e4598")

        .. code-block:: python

            metadata = {
                client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                client.export_assets.ConfigurationMetaNames.ALL_ASSETS: True}

            details = client.export_assets.start(meta_props=metadata, space_id="98a53931-a8c0-4c2f-8319-c793155e4598")

            """
        WMLResource._chk_and_block_create_update_for_python36(self)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Either 'space_id' or 'project_id' has to be provided")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        Export._validate_type(meta_props, u'meta_props', dict, True)
        self._validate_input_meta(meta_props)

        meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        start_meta = {}
        assets = {}

        start_meta[u'name'] = meta[u'name']
        if 'description' in meta:
            start_meta[u'description'] = meta[u'description']

        if "all_assets" not in meta:
            assets.update({"all_assets": False})
        else:
            assets.update({"all_assets": meta[u'all_assets']})

        if "asset_types" in meta:
            assets.update({"asset_types": meta[u'asset_types']})

        if "asset_ids" in meta:
            assets.update(({"asset_ids": meta[u'asset_ids']}))

        start_meta[u'assets'] = assets

        href = self._client.service_instance._href_definitions.exports_href()

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        creation_response = requests.post(href,
                                          params=params,
                                          headers=self._client._get_headers(),
                                          json=start_meta)

        details = self._handle_response(expected_status_code=202,
                                        operationName=u'export start',
                                        response=creation_response)

        export_id = details[u'metadata']['id']

        print("export job with id {} has started. Monitor status using client.export_assets.get_details api. "
              "Check 'help(client.export_assets.get_details)' for details on the api usage".format(export_id))

        return details

    def _validate_input_meta(self, meta_props):
        if 'name' not in meta_props:
            raise WMLClientError("Its mandatory to provide 'NAME' in meta_props. Example: "
                                 "client.export_assets.ConfigurationMetaNames.NAME: 'name'")

        if 'all_assets' not in meta_props and 'asset_ids' not in meta_props and 'asset_types' not in meta_props:
            raise WMLClientError("Its mandatory to provide either 'ALL_ASSETS' or 'ASSET_IDS' or 'ASSET_TYPES' " 
                                 "in meta_props. Example: client.export_assets.ConfigurationMetaNames.ALL_ASSETS: True")

        count = 0

        if 'all_assets' in meta_props:
            count = count + 1
        if 'asset_ids' in meta_props:
            count = count + 1
        if 'asset_types' in meta_props:
            count = count + 1

        if count > 1:
            raise WMLClientError("Only one of 'ALL_ASSETS' or 'ASSET_IDS' or 'ASSET_TYPES' can be provided")

    def cancel(self, export_id, space_id=None, project_id=None):
        """Cancel an export job. `space_id` or `project_id` has to be provided.

        .. note::
            To delete a `export_id` job, use ``delete()`` api.

        :param export_id: export job identifier
        :type export_id: str
        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project_id: str, optional

        :returns: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.export_assets.cancel(export_id='6213cf1-252f-424b-b52d-5cdd9814956c',
                                        space_id='3421cf1-252f-424b-b52d-5cdd981495fe')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        Export._validate_type(export_id, u'export_id', str, True)

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        href = self._client.service_instance._href_definitions.export_href(export_id)

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        cancel_response = requests.delete(href,
                                          params=params,
                                          headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=204,
                                        operationName=u'cancel export',
                                        response=cancel_response)

        if "SUCCESS" == details:
            print("Export job cancelled")

    def delete(self, export_id, space_id=None, project_id=None):
        """Deletes the given `export_id` job. `space_id` or `project_id` has to be provided.

        :param export_id: export job identifier
        :type export_id: str
        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project_id: str, optional

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.export_assets.delete(export_id='6213cf1-252f-424b-b52d-5cdd9814956c',
                                        space_id= '98a53931-a8c0-4c2f-8319-c793155e4598')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        Export._validate_type(export_id, u'export_id', str, True)

        href = self._client.service_instance._href_definitions.export_href(export_id)

        params = {"hard_delete": True}

        if space_id is not None:
            params.update({'space_id': space_id})
        else:
            params.update({'project_id': project_id})

        delete_response = requests.delete(href,
                                          params=params,
                                          headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=204,
                                        operationName=u'delete export job',
                                        response=delete_response)

        if "SUCCESS" == details:
            print("Export job deleted")

    def get_details(self, export_id=None, space_id=None, project_id=None, limit=None, asynchronous=False,
                    get_all=False):
        """Get metadata of the given export job. if no `export_id` is specified all exports metadata is returned.

        :param export_id: export job identifier
        :type export_id: str, optional
        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project_id: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional
        :param asynchronous: if `True`, it will work as a generator
        :type asynchronous: bool, optional
        :param get_all: if `True`, it will get all entries in 'limited' chunks
        :type get_all: bool, optional

        :return: export(s) metadata
        :rtype: dict (if export_id is not None) or {"resources": [dict]} (if export_id is None)

        **Example**

        .. code-block:: python

            details = client.export_assets.get_details(export_id, space_id= '98a53931-a8c0-4c2f-8319-c793155e4598')
            details = client.export_assets.get_details()
            details = client.export_assets.get_details(limit=100)
            details = client.export_assets.get_details(limit=100, get_all=True)
            details = []
            for entry in client.export_assets.get_details(limit=100, asynchronous=True, get_all=True):
                details.extend(entry)
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")
        
        Export._validate_type(export_id, u'export_id', str, False)
        Export._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.exports_href()

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        if export_id is None:
            return self._get_artifact_details(href, export_id, limit, 'export job', query_params=params,
                                              _async=asynchronous, _all=get_all)

        else:
            return self._get_artifact_details(href, export_id, limit, 'export job', query_params=params)
    
    def list(self, space_id=None, project_id=None, limit=None):
        """Print export jobs in a table format. If limit is set to None there will be only first 50 records shown.

        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project_id: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.export_assets.list()
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        if space_id is not None:
            resources = self.get_details(space_id=space_id)[u'resources']
        else:
            resources = self.get_details(project_id=project_id)[u'resources']


        values = [(m[u'metadata'][u'id'],
                   m[u'metadata'][u'name'],
                   m[u'metadata'][u'created_at'],
                   m[u'entity'][u'status'][u'state']) for m in resources]

        self._list(values, [u'ID', u'NAME', u'CREATED', u'STATUS'], limit, _DEFAULT_LIST_LENGTH)

    @staticmethod
    def get_id(export_details):
        """Get ID of export job from export details.

        :param export_details: metadata of the export job
        :type export_details: dict

        :return: ID of the export job
        :rtype: str

        **Example**

        .. code-block:: python

            id = client.export_assets.get_id(export_details)
        """
        Export._validate_type(export_details, u'export_details', object, True)

        return WMLResource._get_required_element_from_dict(export_details,
                                                           u'export_details',
                                                           [u'metadata', u'id'])

    def get_exported_content(self, export_id, space_id=None, project_id=None, file_path=None):
        """Get the exported content as a zip file.

        :param export_id: export job identifier
        :type export_id: str
        :param space_id: space identifier
        :type space_id: str, optional
        :param project_id: project identifier
        :type project_id: str, optional
        :param file_path: name of local file to create, this should be absolute path of the file
            and the file shouldn't exist
        :type file_path: str, optional


        :return: path to the downloaded function content
        :rtype: str

        **Example**

        .. code-block:: python

            client.exports.get_exported_content(export_id,
                                                space_id='98a53931-a8c0-4c2f-8319-c793155e4598',
                                                file_path='/home/user/my_exported_content.zip')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        if os.path.isfile(file_path):
            raise WMLClientError(u'File with name: \'{}\' already exists.'.format(file_path))

        Export._validate_type(file_path, u'file_path', str, True)

        href = self._client.service_instance._href_definitions.export_content_href(export_id)

        try:
            response = requests.get(href,
                                    params=params,
                                    headers=self._client._get_headers(),
                                    stream=True)

            if response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading export content"), response)

            downloaded_exported_content = response.content
            self._logger.info(u'Successfully downloaded artifact with artifact_url: {}'.format(href))
        except WMLClientError as e:
            raise e
        except Exception as e:
            raise WMLClientError(u'Downloading export content with artifact_url: \'{}\' failed.'.format(href), e)

        try:
            with open(file_path, 'wb') as f:
                f.write(downloaded_exported_content)
            print(u'Successfully saved export content to file: \'{}\''.format(file_path))
            return file_path
        except IOError as e:
            raise WMLClientError(u'Downloading export content with artifact_url: \'{}\' failed.'.format(href), e)




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import importlib
if importlib.util.find_spec("pyhelayers") is not None:
    import pyhelayers as pyhe
from ibmfl.crypto.infra.crypto_he_int import CryptoHe
from ibmfl.crypto.crypto_exceptions import *

class CryptoHeLy(CryptoHe):
    """
    This class implements the interface for HE keys generation functions using pyhelayers. 
    """

    CONFIG_FIELD_SEC_PARAM = "sec_param"
    CONFIG_FIELD_INT_PRECISION = "integer_precision"
    CONFIG_FIELD_FRAC_PRECISION = "fractional_precision"
    CONFIG_FIELD_MULT_DEPTH = "multiplication_depth"
    CONFIG_FIELD_NUM_SLOTS = "num_slots"
    DEFAULT_CONFIG = {
        CONFIG_FIELD_SEC_PARAM: int(128),
        CONFIG_FIELD_INT_PRECISION: int(10),
        CONFIG_FIELD_FRAC_PRECISION: int(29),
        CONFIG_FIELD_MULT_DEPTH: int(1),
        CONFIG_FIELD_NUM_SLOTS: int(2048)
    }

    def __init__(self, config = None):
        super(CryptoHeLy, self).__init__(config)
        if config is None:
            config = CryptoHeLy.DEFAULT_CONFIG
        elif CryptoHeLy.CONFIG_FIELD_SEC_PARAM not in config or \
            CryptoHeLy.CONFIG_FIELD_INT_PRECISION not in config or \
            CryptoHeLy.CONFIG_FIELD_FRAC_PRECISION not in config or \
            CryptoHeLy.CONFIG_FIELD_MULT_DEPTH not in config or \
            CryptoHeLy.CONFIG_FIELD_NUM_SLOTS not in config:
            raise KeyDistributionInputException("Incomplete config=" + repr(config))
        self.sec_param = config[CryptoHeLy.CONFIG_FIELD_SEC_PARAM]
        self.integer_precision = config[CryptoHeLy.CONFIG_FIELD_INT_PRECISION]
        self.fractional_precision = config[CryptoHeLy.CONFIG_FIELD_FRAC_PRECISION]
        self.multiplication_depth = config[CryptoHeLy.CONFIG_FIELD_MULT_DEPTH]
        self.num_slots = config[CryptoHeLy.CONFIG_FIELD_NUM_SLOTS]
        self.context = None
        self.buffer_ctx = None
        self.buffer_sk = None

    def generate_keys(self):
        requirement = pyhe.HeConfigRequirement(
            num_slots = self.num_slots,
            multiplication_depth = self.multiplication_depth,
            fractional_part_precision = self.fractional_precision,
            integer_part_precision = self.integer_precision,
            security_level = self.sec_param)
        self.context = pyhe.DefaultContext()
        self.context.init(requirement)
        self.buffer_ctx = self.context.save_to_buffer()
        self.buffer_sk = self.context.save_secret_key()

    def get_public_key(self):
        if self.buffer_ctx is None:
            raise KeyDistributionInputException("self.buffer_ctx is None")
        return self.buffer_ctx

    def get_private_key(self):
        if self.buffer_sk is None:
            raise KeyDistributionInputException("self.buffer_sk is None")
        return self.buffer_sk



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import os
import logging
import logging.config
import time

from ibmfl._version import __version__
from ibmfl.exceptions import InvalidConfigurationException

logger = logging.getLogger(__name__)


class UTCFormatter(logging.Formatter):
    converter = time.gmtime

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            if "%F" in datefmt:
                msec = "%03d" % record.msecs
                datefmt = datefmt.replace("%F", msec)
            s = time.strftime(datefmt, ct)
        else:
            t = time.strftime("%Y-%m-%dT%H:%M:%S.%F", ct)
            s = "%s.%03dZ" % (t, record.msecs)
        return s


class FLVersionFilter(logging.Filter):

    def filter(self, record):
        record.version = __version__
        return True


def configure_file_logging(filename, log_level='INFO'):
    log_level = logging.getLevelName(log_level)

    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "fl_std": {
                "()": UTCFormatter,
                "format": "%(asctime)s | %(version)s | %(levelname)s | %(name)-50s | %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S.%FZ"
            }
        },
        "handlers": {
            "WriteFile": {
                "class": "logging.FileHandler",
                "filters": ['version_filter'],
                "level": "DEBUG",
                "formatter": "fl_std",
                "filename": filename
            }

        },
        "loggers": {
            "ibmfl": {
                "level": log_level,
                "handlers": [
                    "WriteFile"
                ],
                "propagate": False
            }
        }
    }

    result = configure_logging(dict_config, log_level)
    return result


def configure_logging(config, log_level=logging.INFO):
    """
    configures logging for application based on the configuration file

    :param config: yaml file containing the definitions of formatter and handler
    :type config: `dict`
    :param log_level: should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]
            based on the required granularity
    :type log_level: `int`
    :return: a boolean object. False for default basic config True otherwise
    :rtype: `boolean`
    """

    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "fl_std": {
                "()": UTCFormatter,
                "format": "%(asctime)s | %(version)s | %(levelname)s | %(name)-50s | %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S.%FZ"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "filters": ['version_filter'],
                "level": log_level,
                "formatter": "fl_std",
                "stream": "ext://sys.stdout"
            }

        },
        "loggers": {
            "ibmfl": {
                "level": log_level,
                "handlers": [
                    "console"
                ],
                "propagate": False
            }
        },
        "root": {
            "level": log_level,
            "handlers": [
                "console"
            ]
        }
    }
    if not config:
        config = dict_config

    log_filters = {
        'version_filter': {
            '()': FLVersionFilter
        }
    }
    if 'filters' in config:
        config['filters']['version_filter'] = log_filters['version_filter']
    else:
        config['filters'] = log_filters

    add_version_filter(config)
    result = False
    if config:
        try:
            logging.config.dictConfig(config)
            result = True
        except InvalidConfigurationException as ice:
            logging.error('Failed to load log configuration %s', ice)
            configure_basic_logging(log_level=log_level)
    else:
        configure_basic_logging(log_level=log_level)

    return result


def add_version_filter(config):
    """
    Add versioning filter to log config if not provided

    :param config: yaml file containing the definitions of formatter and handler
    :type config: `dict`
    :return: None
    """
    filters = ['version_filter']
    if 'handlers' in config:

        handlers = config['handlers']

        for key in handlers:
            handler = handlers.get(key)

            if 'filters' in handler:
                filt = handler.get('filters')
                if 'version_filter' not in filt:
                    filt.append('version_filter')
            else:
                handler['filters'] = filters


def configure_basic_logging(log_level=logging.INFO):
    """
    configures logging for the session based on pre defined format.
    default logging will be done on console.

    :param level: should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]
            based on the required granularity
    :type log_level: `int`
    :return: None
    """

    logging.basicConfig(level=log_level,
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # logging.config.dictConfig(default_log_config)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
from cryptography.fernet import Fernet
from ibmfl.crypto.infra.crypto_sym_int import CryptoSym


class CryptoSymFernet(CryptoSym):
    """
    This class implements the interface for symmetric encryption functions using Fernet.
    """

    def __init__(self, key: bytes = None, **kwargs):
        super(CryptoSymFernet, self).__init__(key)
        if key is None:
            self.generate_key()
        else:
            self.key = key
            self.cipher = Fernet(self.key)

    def generate_key(self):
        self.key = Fernet.generate_key()
        self.cipher = Fernet(self.key)
        return



"""
Module providing utility functions helpful for preproccessing data
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import numpy as np
import pandas as pd
from sklearn import preprocessing
from diffprivlib import tools as dp
import logging

from ibmfl.exceptions import FLException

logger = logging.getLogger(__name__)

# TODO get dp stats


def get_min(data, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and
    each column is one feature.
    :type data: `np.ndarray`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the minimum value of each feature across all samples, \
    e.g., epsilon and delta, etc. \
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the minimum value \
    of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        min_vec = np.min(data, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the minimum value. ' + str(ex))
    return min_vec


def get_max(data, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type `np.ndarray`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the maximum value of each feature across all samples, \
    e.g., epsilon and delta, etc.
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the maximum value \
    of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        max_vec = np.max(data, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the maximum value. ' + str(ex))
    return max_vec


def get_mean(data, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the maximum value of each feature across all samples, \
    e.g., epsilon and delta, etc.
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the maximum value \
    of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        var_vec = np.var(data, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the mean value. ' + str(ex))
    return var_vec


def get_var(data, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the variance of each feature across all samples, \
    e.g., epsilon, etc.
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the variance
    of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        var_vec = np.var(data, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the variance. ' + str(ex))
    return var_vec


def get_std(data, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the standard deviation of each feature across all samples, \
    e.g., epsilon, etc.
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the
    standard deviation of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        std_vec = np.std(data, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the standard deviation. ' + str(ex))
    return std_vec


def get_quantile(data, percentage, **kwargs):
    """
    Assuming the dataset is loaded as type `np.array`, and has shape
     (num_samples, num_features).

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param percentage: Quantile or sequence of quantiles to compute, \
    which must be between 0 and 1 inclusive.
    :type percentage: `float` or `np.array` of `float`
    :param kwargs: Dictionary of differential privacy arguments \
    for computing the specified quantile of each feature across all samples, \
    e.g., epsilon, etc.
    :type kwargs: `dict`
    :return: A vector of shape (1, num_features) stores the
    standard deviation of each feature across all samples.
    :rtype: `np.array` of `float`
    """
    try:
        quantile_vec = np.quantile(data, q=percentage, axis=0)
    except Exception as ex:
        raise FLException('Error occurred when calculating '
                          'the quantile. ' + str(ex))
    return quantile_vec


def get_normalizer(data, norm='l2'):
    """
    Obtain the normalizer that perform the normalization preprocessing
    technique across all features via sklearn.preprocessing.normalizer API.
    A normalizer will scale a dataset w.r.t. features to unit norm.

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param norm: The norm to use to normalize each non zero sample.
    By default, norm is set to `l2`.
    :type norm: `str`
    :return: The normalizer preprocessor that can be applied to perform \
    normalizing preprocessing step for the party's local dataset \
    via `transform` method.
    :rtype: `sklearn.preprocessing.data.Normalizer`
    """
    try:
        normalizer = preprocessing.Normalizer(norm=norm).fit(data)

        # test the normalizer
        normalizer.transform(data)
    except Exception as ex:
        raise FLException('Error occurred when obtaining '
                          'the normalizer. ' + str(ex))
    return normalizer


def get_standardscaler(data, mean_val=None, std=None):
    """
    Obtain the StandardScaler that perform the standardization preprocessing
    technique with provided mean and standard deviation values
    via sklearn.preprocessing.scaler API.
    A StandardScaler will standardize a dataset along any axis, and
    center to the mean and component wise scale to unit variance.

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param mean_val: (Optional) A vector of mean values one wants to scale \
    the dataset. \
    The vector should be of shape (1, num_features).
    :type mean_val: `np.ndarray`
    :param std: (Optional) A vector of standard deviation values \
    one wants to scale the dataset. \
    The vector should be of shape (1, num_features).
    :type std: `np.ndarray`
    :return: The standard scaler preprocessor that can be applied to perform \
    standardization preprocessing step for the party's local dataset \
    via `transform` method.
    :rtype: 'sklearn.preprocessing.data.StandardScaler'
    """
    try:
        scaler = preprocessing.StandardScaler().fit(data)

        # set scaler with correct mean_val and std values
        if mean_val is not None:
            logger.info("Set mean_val value of the StandardScaler "
                        "as the provided mean...")
            scaler.mean_ = mean_val
        if std is not None:
            logger.info("Set standard deviation of the StandardScaler "
                        "as the provided standard deviation...")
            scaler.scale_ = std
    except Exception as ex:
        raise FLException('Error occurred when obtaining '
                          'the standardscaler. ' + str(ex))

    return scaler


def get_minmaxscaler(data, feature_range=(0, 1)):
    """
    Obtain a MinMaxScaler that perform the MinMaxScale preprocessing technique
     with provided feature range via sklearn.preprocessing.minmax_scale API.

    A MinMaxScaler will transforms features by scaling each feature to
    a given range.
    This estimator scales and translates each feature individually
    such that it is in the given range on the training set,
    i.e. between zero and one.

    :param data: Provided dataset, assume each row is a data sample and \
    each column is one feature.
    :type data: `np.ndarray`
    :param feature_range: Desired range of transformed data.
    :type feature_range: tuple (min, max), default=(0, 1)
    :return: The minmaxscaler preprocessor that can be applied to perform \
    minmax scaling preprocessing step for the party's local dataset \
    via `transform` method.
    :rtype: `sklearn.preprocessing.data.MinMaxScaler`
    """
    try:
        scaler = preprocessing.MinMaxScaler(feature_range=feature_range).\
            fit(data)
    except Exception as ex:
        raise FLException('Error occurred when obtaining '
                          'the minmaxcaler. ' + str(ex))
    return scaler


def get_reweighing_weights(data, sensitive_attribute, columns):
    """
    Calculates reweighing weights for points, assuming:
    * privileged group has sensitive attribute value = 1, unprivileged group is 0
    * positive class has value = 1, negative class is 0
    weight = P_expected(sensitive_attribute & class)/P_observed(sensitive_attribute & class)

    :param data: Provided dataset, assume each row is a data sample and
    each column is one feature.
    :type data: `tuple`
    :param sensitive_attribute: Sensitive attribute
    :type sensitive_attribute: `str`
    :param columns: dataset column names
    :type columns: `list`
    :return: weights
    :rtype: `np.array`
    """
    (features, labels) = data

    training_dataset = pd.DataFrame(data=features)
    class_values = labels.tolist()

    training_dataset.columns = columns
    training_dataset['class'] = class_values

    nrows = training_dataset.shape[0]

    priv = sum(training_dataset[sensitive_attribute])/nrows
    unpriv = nrows - sum(training_dataset[sensitive_attribute])/nrows
    pos = sum(training_dataset['class'])/nrows
    neg = nrows - sum(training_dataset['class'])/nrows

    tmp_unp_train_data = training_dataset[training_dataset[sensitive_attribute] == 0]
    tmp_p_train_data = training_dataset[training_dataset[sensitive_attribute] == 1]

    if len(tmp_unp_train_data) > 0:
        unpriv_neg = tmp_unp_train_data['class'].value_counts()[0]/nrows
        if sum(tmp_unp_train_data['class']) > 0:
            unpriv_pos = tmp_unp_train_data['class'].value_counts()[1]/nrows
        else:
            unpriv_pos = 0
    else:
        unpriv_neg = 0
    if len(tmp_p_train_data) > 0:
        priv_neg = tmp_p_train_data['class'].value_counts()[0]/nrows
        if sum(tmp_p_train_data['class']) > 0:
            priv_pos = tmp_p_train_data['class'].value_counts()[1]/nrows
        else:
            priv_pos = 0
    else:
        priv_pos = 0

    weight = []
    for index, row in training_dataset.iterrows():
        if row[sensitive_attribute] == 0 and row['class'] == 0:
            weight.append(unpriv * neg / unpriv_neg)
        elif row[sensitive_attribute] == 0 and row['class'] == 1:
            weight.append(unpriv * pos / unpriv_pos)
        elif row[sensitive_attribute] == 1 and row['class'] == 0:
            weight.append(priv * neg / priv_neg)
        elif row[sensitive_attribute] == 1 and row['class'] == 1:
            weight.append(priv * pos / priv_pos)

    return np.array(weight)

def get_hist_counts(data, sensitive_attribute, columns, eps):
    """
    Calculates noisy counts for reweighing with differential privacy (epsilon set in
    datahandler), assuming:
    * privileged group has sensitive attribute value = 1, unprivileged group is 0
    * positive class has value = 1, negative class is 0
    weight = P_expected(sensitive_attribute & class)/P_observed(sensitive_attribute & class)

    :param data: Provided dataset, assume each row is a data sample and
    each column is one feature.
    :type data: `tuple` of `nd.array`
    :param sensitive_attribute: Sensitive attribute
    :type sensitive_attribute: `str`
    :param columns: dataset column names
    :type columns: `list`
    :param eps: epsilon utilized for differential privacy
    :type eps: `float`
    :return: counts
    :rtype: `dict`
    """
    (features, labels) = data
    counts = {}

    training_dataset = pd.DataFrame(data=features)
    class_values = labels.tolist()

    training_dataset.columns = columns
    training_dataset['class'] = class_values

    data_whole = training_dataset[[sensitive_attribute, 'class']]
    data_whole_0 = data_whole[data_whole[sensitive_attribute] == 0]
    data_whole_0_t = data_whole_0.T

    data_whole_0_t = np.array(data_whole_0_t)
    dp_hist_0, dp_bins_0 = dp.histogram(data_whole_0_t[1], bins=2, epsilon=eps)
    counts['unp_neg'] = dp_hist_0[0]
    counts['unp_pos'] = dp_hist_0[1]

    data_whole_1 = data_whole[data_whole[sensitive_attribute] == 1]
    data_whole_1_t = data_whole_1.T

    data_whole_1_t = np.array(data_whole_1_t)
    dp_hist_1, dp_bins_1 = dp.histogram(data_whole_1_t[1], bins=2, epsilon=eps)
    counts['p_neg'] = dp_hist_1[0]
    counts['p_pos'] = dp_hist_1[1]

    tmp_unp_train_data = training_dataset[training_dataset[sensitive_attribute] == 0]
    tmp_p_train_data = training_dataset[training_dataset[sensitive_attribute] == 1]
    unpriv_neg = tmp_unp_train_data['class'].value_counts()[0]
    unpriv_pos = tmp_unp_train_data['class'].value_counts()[1]
    priv_neg = tmp_p_train_data['class'].value_counts()[0]
    priv_pos = tmp_p_train_data['class'].value_counts()[1]

    counts['unpriv_neg'] = unpriv_neg
    counts['unpriv_pos'] = unpriv_pos
    counts['priv_neg'] = priv_neg
    counts['priv_pos'] = priv_pos

    return counts



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""

from __future__ import print_function
import time
import pickle
import logging

from sklearn.utils.validation import check_is_fitted
from sklearn.ensemble._hist_gradient_boosting.loss import *
from sklearn.ensemble._hist_gradient_boosting.common import *
from sklearn.preprocessing import OrdinalEncoder, LabelEncoder
from sklearn.utils.multiclass import check_classification_targets
from sklearn.metrics import check_scoring, accuracy_score, r2_score
from sklearn.utils import check_X_y, check_random_state, check_array
from sklearn.ensemble._hist_gradient_boosting.grower import TreePredictor

from ibmfl.util import fl_metrics
from ibmfl.util import config
from ibmfl.model.fl_model import FLModel
from ibmfl.util.xgboost.utils import is_classifier
from ibmfl.util.xgboost.export import export_sklearn
from ibmfl.exceptions import ModelInitializationException, ModelException
from ibmfl.util.xgboost.hyperparams import init_parameters, \
    validate_parameters

logger = logging.getLogger(__name__)


class XGBFLModel(FLModel, ABC):
    """
    Wrapper class implementation for XGBoost containing the XGBoost Model Object
    """

    def __init__(self, model_name, model_spec, xgb_model=None, **kwargs):
        """
        Create a XGBFLModel instance for XGBoost model based either on an
        existing model object or an entirely new model object.

        :param model_type: String specifying the name of the model
        :type model_type: `str`
        :param model_spec: Hyperparameters associated with the model.
        :type model_spec: `dict`
        :param xgb_model: List of predictors existing predictor structures.
        :type xgb_model: `list`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a XGBoost model.
        :type kwargs: `dict`
        """
        super().__init__(model_name, model_spec, **kwargs)

        # Initialize Input Model Object Parameters
        # Note: For model loading process, onboard from `xgb_model` parameter.
        self._predictors = []

        # Initialize Additional Internal Model Parameters
        self.model_type = 'XGBFLModel'
        self._baseline_prediction = None
        self._raw_predictions = None
        self.n_features_ = 0
        self.loss_ = None

        # Validate and Initialize Model Hyperparameters
        validate_parameters(model_spec)
        init_parameters(self, model_spec)

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        This function is not implemented as model training is not operated
        within the FL Model object and is part of an external process.

        :return: `NotImplementedError`
        """
        return NotImplementedError

    def update_model(self, model_update=None, **kwargs):
        """
        Updates model using provided `model_update`. Additional arguments
        specific to the model can be added through `**kwargs`

        :param model_update: Model with update. This is specific to each model \
        type e.g., `ModelUpdateSGD`. The specific type should be checked by \
        the corresponding FLModel class.
        :type `ModelUpdate`
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: None
        """
        if model_update is not None:
            if isinstance(model_update.get('xgb_model'), list):
                # Perform Model Update
                self._predictors.append(model_update.get('xgb_model'))
            else:
                raise ValueError('Provided model object is not of the correct '
                                 'data type. Should be a `list`. '
                                 'Type provided: ' + str(type(model_update)))

    def _raw_predict(self, x):
        """
        Internal helper function for generating predictions for the model.

        :param x: The input data for the prediction model.
        :type x: `np.array`
        :param model: Model object which is used for generating the prediction. \
        If not provided the default model assigned internally will be used.
        :type model: `XGBFLModel`
        :return: raw_predictions
        :rtype: `np.array`
        """
        logger.info('Performing Model Inference Process')

        # Initialize Prediction Object
        n_samples = x.shape[0]
        preds = np.zeros(shape=(self.n_trees, n_samples))

        # Append Null Model Predictions
        if self._baseline_prediction is not None:
            preds += self._baseline_prediction
        elif self._raw_predictions is not None:
            preds += self._raw_predictions

        # Generate Predictions
        for p in self._predictors:
            for k, p_i in enumerate(p):
                preds[k, :] += p_i.predict(x.astype(np.float64))

        return preds

    def get_model_update(self):
        """
        Since we are not using the conventional train() function at the local
        training handler, we respectively do not implement anything within here,
        nor this function is utilized anywhere else.
        """
        return NotImplementedError

    @abstractmethod
    def predict(self, x):
        """
        Given a set of inputs, generate inference for the given set of samples.

        :param x: The input data for the prediction model.
        :type x: `np.array`
        :param model: Model object which is used for generating the prediction. \
        If not provided the default model assigned internally will be used.
        :type model: `XGBFLModel`
        :return: `np.array`
        """
        return NotImplementedError

    def save_model(self, filename=None, path=None):
        """
        Saves a sklearn model to file in the format specific
        to the framework requirement as a pickle file (only for inference use).

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in the default data location of \
        the library `DATA_PATH`.
        :type path: `str`
        :return: filename
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            filename = '{}_{}.pickle'.format(file, time.time())            
            
        full_path = super().get_model_absolute_path(filename)

        # Scikit-Learn Export Demo
        export_sklearn(self, full_path)

        if len(self._predictors) > 0:
            logger.info('Model saved in path: %s.', full_path)
        else:
            logger.info('Persisting empty model in path: %s.', full_path)

        return filename

    @staticmethod
    def load_model(filename):
        """
        Load model from provided filename

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :param path: Path of the folder where to store the model. If no path \
        is specified, the model will be stored in the default data location of \
        the library `DATA_PATH`.
        :type path: `str`
        :return: Returns the corresponding model object.
        :rtype: `XGBFLModel`
        """
        absolute_path = config.get_absolute_path(filename)

        with open(absolute_path, 'rb') as f:
            model = pickle.load(f)
            self = model

        if len(self._predictors) == 0:
            logger.info('Model does not contain any predictors.')

        return model

    @abstractmethod
    def get_loss(self):
        """
        Internal helper function used to obtain the corresponding loss function
        which is dependent on the learning task set by the hyperparameters.

        :return: Returns the loss object used to compute the loss function.
        :rtype: `BaseLoss` based object
        """
        return NotImplementedError

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, test) or a datagenerator of of type `keras.utils.Sequence`, \
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`
        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test)

        else:
            raise ModelException("Invalid test dataset!")

    @abstractmethod
    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates model given the samples x and true labels y.
        Multiple evaluation metrics are returned in a dictionary

        :param x: Samples with shape as expected by the model.
        :type x: Data structure as expected by the model \
        :param y: Corresponding labels to x
        :type y: Data structure the same as the type defines labels \
        in testing data.
        :param batch_size: Size of batches.
        :type batch_size: `int`
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: Dictionary with all evaluation metrics provided by specific \
        implementation.
        :rtype: `dict`
        """
        return NotImplementedError


class XGBRegressorFLModel(XGBFLModel):
    _VALID_LOSSES = ('least_squares', 'least_absolute_deviation')

    def predict(self, x):
        """
        Perform prediction for a batch of inputs.

        :param x: Samples with shape as expected by the model.
        :type x: Data structure as expected by the model
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: Predictions
        :rtype: Data structure the same as the type defines labels
        in testing data.
        """
        if len(self._predictors) > 0:
            return self._raw_predict(x).ravel()
        else:
            raise Exception('Model has not been trained yet.')

    def encode_target(self, y):
        """
        Converts the input y to the expected dtype.

        :param y: The corresponding target data from the dataset to encode.
        :type y: `np.array`
        :return: Returns the corresponding encoded y values.
        :rtype: `np.array`
        """
        self.n_trees = 1
        return y.astype(Y_DTYPE, copy=False)

    def get_loss(self, sample_weight):
        """
        Given the initialized loss type defined under the hyerparameters, we
        return the corresponding loss function to dictate the corresponding
        learning task of the model.

        :param sample_weight: Weights of training data
        :type sample_weight: `np.ndarray`
        :return: Returns the respective loss object as defined in the FL \
        hyperparameters.
        :rtype: Derivation of `BaseLoss`
        """
        return _LOSSES[self.loss](sample_weight=sample_weight)

    def evaluate_model(self, x, y, **kwargs):
        """
        Given an input set of values and their values, generate the prediction
        and compute the R^2 metric.

        :param x: The input data for the prediction model.
        :type x: `np.array`
        :param y: The corresponding target value of the prediction.
        :type y: `np.array`
        :return: Return a dictionary containing the R^2 metric for the provided
        input and target values.
        :rtype: `dict`
        """
        metrics = {}
        if len(self._predictors) > 0:
            y_hat = self.predict(x)
            score = r2_score(y, y_hat)
            metrics['r2_score'] = score
            additional_metrics = fl_metrics.get_eval_metrics_for_regression(
                y, y_hat)
            metrics = {**metrics, **additional_metrics}
            return metrics
        else:
            logger.info('Model has not been trained yet.')
            return {}


class XGBClassifierFLModel(XGBFLModel):
    _VALID_LOSSES = ('binary_crossentropy', 'categorical_crossentropy', 'auto')

    def encode_target(self, y):
        """
        Converts the input y to the expected dtype and performs a label
        encoding. Here, we assume that each party has at least one sample of
        the corresponding class label type for each different classes.

        :param y: The corresponding target data from the dataset to encode.
        :type  y: `np.array`
        :return y: Returns the corresponding encoded y values.
        :rtype  y: `np.array`
        """
        # Validate Classification Target Values
        check_classification_targets(y)

        # Apply Label Encoder Transformation
        lab_enc = LabelEncoder()
        enc_y = lab_enc.fit_transform(y).astype(np.float64, copy=False)

        # Extract Encoded Target Sizes
        self.classes_ = lab_enc.classes_
        if self.classes_.shape[0] != self.num_classes:
            raise ValueError('Number of classes defined in configuration file '
                             'and the classes derived from the data does not '
                             'match. Found {0} classes, while config file '
                             'is defined as {1} classes.'.format(
                             self.classes_.shape[0], self.num_classes))

        if self.loss == 'auto':
            self.n_trees = 1 if self.classes_.shape[0] <= 2 else self.classes_.shape[0]
        else:
            self.n_trees = 1 if self.num_classes <= 2 else self.num_classes

        return enc_y

    def get_loss(self, sample_weight):
        """
        Given the initialized loss type defined under the hyerparameters, we
        return the corresponding loss function to dictate the corresponding
        learning task of the model. If auto is selected, then we will
        automatically determine whether the classification task is binary or
        multiclass given the label encoding cardinality.

        :param sample_weight: Weights of training data
        :type sample_weight: `np.ndarray`
        :return: Returns the respective loss object as defined in the FL \
        hyperparameters.
        :rtype: Derivation of `BaseLoss`
        """
        if (self.loss == 'categorical_crossentropy' and self.n_trees == 1):
            raise ValueError("Incompatible loss and target variable counts.")

        if self.loss == 'auto':
            return _LOSSES['binary_crossentropy']() if self.n_trees == 1 else \
                _LOSSES['categorical_crossentropy']()
        else:
            return _LOSSES[self.loss](sample_weight=sample_weight)

    def predict(self, x, **kwargs):
        """
        Perform prediction for a batch of inputs returned as class values.

        :param x: Samples with shape as expected by the model.
        :type x: Data structure as expected by the model
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: Predictions based on class values.
        :rtype: Data structure the same as the type defines labels
        in testing data.
        """
        if len(self._predictors) > 0:
            encoded_classes = np.argmax(self.predict_proba(x), axis=1)
            return self.classes_[encoded_classes]
        else:
            raise Exception('Model has not been trained yet.')

    def predict_proba(self, x, **kwargs):
        """
        Perform prediction for a batch of inputs as probabilities.

        :param x: Samples with shape as expected by the model.
        :type x: Data structure as expected by the model
        :param kwargs: Dictionary of model-specific arguments.
        :type kwargs: `dict`
        :return: A probabilistic set of predictions
        :rtype: Data structure the same as the type defines labels \
        in testing data.
        """
        if len(self._predictors) > 0:
            raw_predictions = self._raw_predict(x)
            return self.loss_.predict_proba(raw_predictions)
        else:
            raise Exception('Model has not been trained yet.')

    def evaluate_model(self, x, y, **kwargs):
        """
        Given an input set of values and their values, generate the prediction
        and compute the accuracy.

        :param x: The input data for the prediction model.
        :type x: `np.array`
        :param y: The corresponding target value of the prediction.
        :type y: `np.array`
        :return: Return a dictionary containing the accuracy for the provided \
        input and target values.
        :rtype: `dict`
        """
        acc = {}
        if len(self._predictors) > 0:
            y_hat = self.predict(x)
            correct = 0
            for i in range(x.shape[0]):
                if y_hat[i] == y[i]:
                    correct += 1

            acc = {'acc': correct/float(len(y))}
            additional_metrics = fl_metrics.get_eval_metrics_for_classificaton(
                y, y_hat)

            acc = {**acc, **additional_metrics}
            return acc
        else:
            logger.info('No models have been trained yet.')
            return {}



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""

from __future__ import print_function


def is_classifier(cls):
    """
    Auxillary function to validate whether a given object is a
    classification-based model. (Can be used for the Fusion Handler as well as
    the Local Training Handler).

    :param cls: Object of interest to identify the type of model.
    :type  cls: `Object`
    :return `bool`
    """
    return cls.loss == 'auto' or cls.loss == 'binary_crossentropy' or \
        cls.loss == 'categorical_crossentropy'
